# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
import pecan

from octavia.api import config as pconfig
from octavia.api.healthcheck import healthcheck_plugins
from octavia.tests.functional.db import base as base_db_test


class TestHealthCheck(base_db_test.OctaviaDBTestBase):

    def setUp(self):
        super(TestHealthCheck, self).setUp()

        # We need to define these early as they are late loaded in oslo
        # middleware and our configuration overrides would not apply.
        # Note: These must match exactly the option definitions in
        # oslo.middleware healthcheck! If not you will get duplicate option
        # errors.
        healthcheck_opts = [
            cfg.BoolOpt(
                'detailed', default=False,
                help='Show more detailed information as part of the response. '
                     'Security note: Enabling this option may expose '
                     'sensitive details about the service being monitored. '
                     'Be sure to verify that it will not violate your '
                     'security policies.'),
            cfg.ListOpt(
                'backends', default=[],
                help='Additional backends that can perform health checks and '
                     'report that information back as part of a request.'),
        ]
        cfg.CONF.register_opts(healthcheck_opts, group='healthcheck')

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.config(group='healthcheck', backends=['octavia_db_check'])
        self.conf.config(group='api_settings', healthcheck_refresh_interval=5)
        self.UNAVAILABLE = (healthcheck_plugins.OctaviaDBHealthcheck.
                            UNAVAILABLE_REASON)

        def reset_pecan():
            pecan.set_config({}, overwrite=True)

        self.addCleanup(reset_pecan)

    def _make_app(self):
        # Note: we need to set argv=() to stop the wsgi setup_app from
        # pulling in the testing tool sys.argv
        return pecan.testing.load_test_app({'app': pconfig.app,
                                            'wsme': pconfig.wsme}, argv=())

    def _get_enabled_app(self):
        self.conf.config(group='api_settings', healthcheck_enabled=True)
        return self._make_app()

    def _get_disabled_app(self):
        self.conf.config(group='api_settings', healthcheck_enabled=False)
        return self._make_app()

    def _get(self, app, path, params=None, headers=None, status=200,
             expect_errors=False):
        response = app.get(path, params=params, headers=headers, status=status,
                           expect_errors=expect_errors)
        return response

    def _head(self, app, path, headers=None, status=204, expect_errors=False):
        response = app.head(path, headers=headers, status=status,
                            expect_errors=expect_errors)
        return response

    def _post(self, app, path, body, headers=None, status=201,
              expect_errors=False):
        response = app.post_json(path, params=body, headers=headers,
                                 status=status, expect_errors=expect_errors)
        return response

    def _put(self, app, path, body, headers=None, status=200,
             expect_errors=False):
        response = app.put_json(path, params=body, headers=headers,
                                status=status, expect_errors=expect_errors)
        return response

    def _delete(self, app, path, params=None, headers=None, status=204,
                expect_errors=False):
        response = app.delete(path, headers=headers, status=status,
                              expect_errors=expect_errors)
        return response

    def test_healthcheck_get_text(self):
        self.conf.config(group='healthcheck', detailed=False)
        response = self._get(self._get_enabled_app(), '/healthcheck')
        self.assertEqual(200, response.status_code)
        self.assertEqual('OK', response.text)

    # Note: For whatever reason, detailed=True text has no additonal info
    def test_healthcheck_get_text_detailed(self):
        self.conf.config(group='healthcheck', detailed=True)
        response = self._get(self._get_enabled_app(), '/healthcheck')
        self.assertEqual(200, response.status_code)
        self.assertEqual('OK', response.text)

    def test_healthcheck_get_json(self):
        self.conf.config(group='healthcheck', detailed=False)
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'application/json'})
        self.assertEqual(200, response.status_code)
        self.assertFalse(response.json['detailed'])
        self.assertEqual(['OK'], response.json['reasons'])

    def test_healthcheck_get_json_detailed(self):
        self.conf.config(group='healthcheck', detailed=True)
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'application/json'})
        self.assertEqual(200, response.status_code)
        self.assertTrue(response.json['detailed'])
        self.assertEqual('OK', response.json['reasons'][0]['reason'])
        self.assertTrue(response.json['gc'])

    def test_healthcheck_get_html(self):
        self.conf.config(group='healthcheck', detailed=False)
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'text/html'})
        self.assertEqual(200, response.status_code)
        self.assertIn('OK', response.text)

    def test_healthcheck_get_html_detailed(self):
        self.conf.config(group='healthcheck', detailed=True)
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'text/html'})
        self.assertEqual(200, response.status_code)
        self.assertIn('OK', response.text)
        self.assertIn('Garbage collector', response.text)

    def test_healthcheck_get_text_cached(self):
        self.conf.config(group='healthcheck', detailed=False)
        app = self._get_enabled_app()
        for i in range(10):
            response = self._get(app, '/healthcheck')
            self.assertEqual(200, response.status_code)
            self.assertEqual('OK', response.text)

    def test_healthcheck_disabled_get(self):
        self._get(self._get_disabled_app(), '/healthcheck', status=404)

    def test_healthcheck_head(self):
        response = self._head(self._get_enabled_app(), '/healthcheck')
        self.assertEqual(204, response.status_code)

    def test_healthcheck_disabled_head(self):
        self._head(self._get_disabled_app(), '/healthcheck', status=404)

    # These should be denied by the API
    def test_healthcheck_post(self):
        self._post(self._get_enabled_app(), '/healthcheck',
                   {'foo': 'bar'}, status=405)

    def test_healthcheck_put(self):
        self._put(self._get_enabled_app(), '/healthcheck',
                  {'foo': 'bar'}, status=405)

    def test_healthcheck_delete(self):
        self._delete(self._get_enabled_app(), '/healthcheck',
                     status=405)

    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_get_failed(self, mock_get_session):
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             status=503)
        self.assertEqual(503, response.status_code)
        self.assertEqual(self.UNAVAILABLE, response.text)

    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_head_failed(self, mock_get_session):
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._head(self._get_enabled_app(), '/healthcheck',
                              status=503)
        self.assertEqual(503, response.status_code)

    @mock.patch('octavia.db.healthcheck.check_database_connection',
                side_effect=Exception('boom'))
    def test_healthcheck_get_failed_check(self, mock_db_check):
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             status=503)
        self.assertEqual(503, response.status_code)
        self.assertEqual(self.UNAVAILABLE, response.text)

    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_get_json_failed(self, mock_get_session):
        self.conf.config(group='healthcheck', detailed=False)
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'application/json'},
                             status=503)
        self.assertEqual(503, response.status_code)
        self.assertFalse(response.json['detailed'])
        self.assertEqual([self.UNAVAILABLE],
                         response.json['reasons'])

    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_get_json_detailed_failed(self, mock_get_session):
        self.conf.config(group='healthcheck', detailed=True)
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'application/json'},
                             status=503)
        self.assertEqual(503, response.status_code)
        self.assertTrue(response.json['detailed'])
        self.assertEqual(self.UNAVAILABLE,
                         response.json['reasons'][0]['reason'])
        self.assertIn('boom', response.json['reasons'][0]['details'])

    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_get_html_failed(self, mock_get_session):
        self.conf.config(group='healthcheck', detailed=False)
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'text/html'}, status=503)
        self.assertEqual(503, response.status_code)
        self.assertIn(self.UNAVAILABLE, response.text)

    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_get_html_detailed_failed(self, mock_get_session):
        self.conf.config(group='healthcheck', detailed=True)
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             headers={'Accept': 'text/html'}, status=503)
        self.assertEqual(503, response.status_code)
        self.assertIn(self.UNAVAILABLE, response.text)
        self.assertIn('boom', response.text)
        self.assertIn('Garbage collector', response.text)

    # Note: For whatever reason, detailed=True text has no additonal info
    @mock.patch('octavia.db.api.get_session')
    def test_healthcheck_get_text_detailed_failed(self, mock_get_session):
        self.conf.config(group='healthcheck', detailed=True)
        mock_session = mock.MagicMock()
        mock_session.execute.side_effect = [Exception('boom')]
        mock_get_session.return_value = mock_session
        response = self._get(self._get_enabled_app(), '/healthcheck',
                             status=503)
        self.assertEqual(503, response.status_code)
        self.assertEqual(self.UNAVAILABLE, response.text)



#    Copyright 2019 Verizon Media
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.common import constants
import octavia.common.context
from octavia.common import exceptions
from octavia.tests.functional.api.v2 import base


class TestAvailabilityZones(base.BaseAPITest):
    root_tag = 'availability_zone'
    root_tag_list = 'availability_zones'
    root_tag_links = 'availability_zones_links'

    def setUp(self):
        super(TestAvailabilityZones, self).setUp()
        self.azp = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')

    def _assert_request_matches_response(self, req, resp, **optionals):
        self.assertNotIn('id', resp)  # AZs do not expose an ID
        req_description = req.get('description')
        self.assertEqual(req.get('name'), resp.get('name'))
        if not req_description:
            self.assertEqual('', resp.get('description'))
        else:
            self.assertEqual(req.get('description'), resp.get('description'))
        self.assertEqual(req.get('availability_zone_profile_id'),
                         resp.get('availability_zone_profile_id'))
        self.assertEqual(req.get('enabled', True),
                         resp.get('enabled'))

    def test_empty_list(self):
        response = self.get(self.AZS_PATH)
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual([], api_list)

    def test_create(self):
        az_json = {'name': 'test1',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        self._assert_request_matches_response(az_json, api_az)

    def test_create_with_missing_name(self):
        az_json = {'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body, status=400)
        err_msg = ("Invalid input for field/attribute name. Value: "
                   "'None'. Mandatory field missing.")
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_with_long_name(self):
        az_json = {'name': 'n' * 256,
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        self.post(self.AZS_PATH, body, status=400)

    def test_create_with_long_description(self):
        az_json = {'name': 'test-az',
                   'description': 'n' * 256,
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        self.post(self.AZS_PATH, body, status=400)

    def test_create_with_missing_availability_zone_profile(self):
        az_json = {'name': 'xyz'}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body, status=400)
        err_msg = (
            "Invalid input for field/attribute availability_zone_profile_id. "
            "Value: 'None'. Mandatory field missing.")
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_with_bad_availability_zone_profile(self):
        az_json = {'name': 'xyz', 'availability_zone_profile_id': 'bogus'}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body, status=400)
        err_msg = (
            "Invalid input for field/attribute availability_zone_profile_id. "
            "Value: 'bogus'. Value should be UUID format")
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_duplicate_names(self):
        self.create_availability_zone(
            'name', 'description', self.azp.get('id'), True)
        az_json = {'name': 'name',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body, status=409)
        err_msg = "A availability zone of name already exists."
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_authorized(self):
        az_json = {'name': 'test1',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self._assert_request_matches_response(az_json, api_az)

    def test_create_not_authorized(self):
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        az_json = {'name': 'name',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body, status=403)
        api_az = response.json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, api_az)

    def test_create_db_failure(self):
        az_json = {'name': 'test1',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        with mock.patch("octavia.db.repositories.AvailabilityZoneRepository."
                        "create") as mock_create:
            mock_create.side_effect = Exception
            self.post(self.AZS_PATH, body, status=500)

    def test_get(self):
        az = self.create_availability_zone(
            'name', 'description', self.azp.get('id'), True)
        response = self.get(
            self.AZ_PATH.format(
                az_name=az.get('name'))).json.get(self.root_tag)
        self.assertEqual('name', response.get('name'))
        self.assertEqual('description', response.get('description'))
        self.assertEqual(az.get('name'), response.get('name'))
        self.assertEqual(self.azp.get('id'),
                         response.get('availability_zone_profile_id'))
        self.assertTrue(response.get('enabled'))

    def test_get_one_fields_filter(self):
        az = self.create_availability_zone(
            'name', 'description', self.azp.get('id'), True)
        response = self.get(
            self.AZ_PATH.format(az_name=az.get('name')), params={
                'fields': ['name', 'availability_zone_profile_id']}
        ).json.get(self.root_tag)
        self.assertEqual(az.get('name'), response.get('name'))
        self.assertEqual(self.azp.get('id'),
                         response.get('availability_zone_profile_id'))
        self.assertIn(u'availability_zone_profile_id', response)
        self.assertNotIn(u'description', response)
        self.assertNotIn(u'enabled', response)

    def test_get_one_deleted_name(self):
        response = self.get(
            self.AZ_PATH.format(az_name=constants.NIL_UUID), status=404)
        self.assertEqual(
            'Availability Zone {} not found.'.format(constants.NIL_UUID),
            response.json.get('faultstring'))

    def test_get_authorized(self):
        az = self.create_availability_zone(
            'name', 'description', self.azp.get('id'), True)

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.get(
                    self.AZ_PATH.format(
                        az_name=az.get('name'))).json.get(self.root_tag)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual('name', response.get('name'))
        self.assertEqual('description', response.get('description'))
        self.assertEqual(self.azp.get('id'),
                         response.get('availability_zone_profile_id'))
        self.assertTrue(response.get('enabled'))

    def test_get_not_authorized(self):
        az = self.create_availability_zone(
            'name', 'description', self.azp.get('id'), True)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        response = self.get(self.AZ_PATH.format(
            az_name=az.get('name')), status=403).json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, response)

    def test_get_all(self):
        self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        ref_az_1 = {
            u'description': u'description', u'enabled': True,
            u'availability_zone_profile_id': self.azp.get('id'),
            u'name': u'name1'}
        self.create_availability_zone(
            'name2', 'description', self.azp.get('id'), True)
        ref_az_2 = {
            u'description': u'description', u'enabled': True,
            u'availability_zone_profile_id': self.azp.get('id'),
            u'name': u'name2'}
        response = self.get(self.AZS_PATH)
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual(2, len(api_list))
        self.assertIn(ref_az_1, api_list)
        self.assertIn(ref_az_2, api_list)

    def test_get_all_fields_filter(self):
        self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        self.create_availability_zone(
            'name2', 'description', self.azp.get('id'), True)
        response = self.get(self.AZS_PATH, params={
            'fields': ['id', 'name']})
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual(2, len(api_list))
        for az in api_list:
            self.assertIn(u'name', az)
            self.assertNotIn(u'availability_zone_profile_id', az)
            self.assertNotIn(u'description', az)
            self.assertNotIn(u'enabled', az)

    def test_get_all_authorized(self):
        self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        self.create_availability_zone(
            'name2', 'description', self.azp.get('id'), True)
        response = self.get(self.AZS_PATH)

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                api_list = response.json.get(self.root_tag_list)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(2, len(api_list))

    def test_get_all_not_authorized(self):
        self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        self.create_availability_zone(
            'name2', 'description', self.azp.get('id'), True)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        response = self.get(self.AZS_PATH, status=403).json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, response)

    def test_update(self):
        az_json = {'name': 'Fancy_Availability_Zone',
                   'description': 'A great az. Pick me!',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        availability_zone_name = api_az.get('name')

        az_json = {'description': 'An even better az. Pick me!',
                   'enabled': False}
        body = self._build_body(az_json)
        self.put(self.AZ_PATH.format(az_name=availability_zone_name), body)

        updated_az = self.get(self.AZ_PATH.format(
            az_name=availability_zone_name)).json.get(self.root_tag)
        self.assertEqual('An even better az. Pick me!',
                         updated_az.get('description'))
        self.assertEqual(availability_zone_name, updated_az.get('name'))
        self.assertEqual(self.azp.get('id'),
                         updated_az.get('availability_zone_profile_id'))
        self.assertFalse(updated_az.get('enabled'))

    def test_update_deleted_name(self):
        update_json = {'description': 'fake_desc'}
        body = self._build_body(update_json)
        response = self.put(
            self.AZ_PATH.format(az_name=constants.NIL_UUID), body,
            status=404)
        self.assertEqual(
            'Availability Zone {} not found.'.format(constants.NIL_UUID),
            response.json.get('faultstring'))

    def test_update_none(self):
        az_json = {'name': 'Fancy_Availability_Zone',
                   'description': 'A great az. Pick me!',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        availability_zone_name = api_az.get('name')

        az_json = {}
        body = self._build_body(az_json)
        self.put(self.AZ_PATH.format(az_name=availability_zone_name), body)

        updated_az = self.get(self.AZ_PATH.format(
            az_name=availability_zone_name)).json.get(self.root_tag)
        self.assertEqual('Fancy_Availability_Zone', updated_az.get('name'))
        self.assertEqual('A great az. Pick me!',
                         updated_az.get('description'))
        self.assertEqual(availability_zone_name, updated_az.get('name'))
        self.assertEqual(self.azp.get('id'),
                         updated_az.get('availability_zone_profile_id'))
        self.assertTrue(updated_az.get('enabled'))

    def test_update_availability_zone_profile_id(self):
        az_json = {'name': 'Fancy_Availability_Zone',
                   'description': 'A great az. Pick me!',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        availability_zone_name = api_az.get('name')

        az_json = {'availability_zone_profile_id': uuidutils.generate_uuid()}
        body = self._build_body(az_json)
        self.put(self.AZ_PATH.format(az_name=availability_zone_name),
                 body, status=400)
        updated_az = self.get(self.AZ_PATH.format(
            az_name=availability_zone_name)).json.get(self.root_tag)
        self.assertEqual(self.azp.get('id'),
                         updated_az.get('availability_zone_profile_id'))

    def test_update_authorized(self):
        az_json = {'name': 'Fancy_Availability_Zone',
                   'description': 'A great az. Pick me!',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        availability_zone_name = api_az.get('name')

        az_json = {'description': 'An even better az. Pick me!',
                   'enabled': False}
        body = self._build_body(az_json)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                self.put(self.AZ_PATH.format(az_name=availability_zone_name),
                         body)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        updated_az = self.get(self.AZ_PATH.format(
            az_name=availability_zone_name)).json.get(self.root_tag)
        self.assertEqual('An even better az. Pick me!',
                         updated_az.get('description'))
        self.assertEqual(availability_zone_name, updated_az.get('name'))
        self.assertEqual(self.azp.get('id'),
                         updated_az.get('availability_zone_profile_id'))
        self.assertFalse(updated_az.get('enabled'))

    def test_update_not_authorized(self):
        az_json = {'name': 'Fancy_Availability_Zone',
                   'description': 'A great az. Pick me!',
                   'availability_zone_profile_id': self.azp.get('id')}
        body = self._build_body(az_json)
        response = self.post(self.AZS_PATH, body)
        api_az = response.json.get(self.root_tag)
        availability_zone_name = api_az.get('name')

        az_json = {'description': 'An even better az. Pick me!',
                   'enabled': False}
        body = self._build_body(az_json)

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        self.put(self.AZ_PATH.format(az_name=availability_zone_name),
                 body, status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        updated_az = self.get(self.AZ_PATH.format(
            az_name=availability_zone_name)).json.get(self.root_tag)
        self.assertEqual('A great az. Pick me!',
                         updated_az.get('description'))
        self.assertEqual(availability_zone_name, updated_az.get('name'))
        self.assertEqual(self.azp.get('id'),
                         updated_az.get('availability_zone_profile_id'))
        self.assertTrue(updated_az.get('enabled'))

    @mock.patch('octavia.db.repositories.AvailabilityZoneRepository.update')
    def test_update_exception(self, mock_update):
        mock_update.side_effect = [exceptions.OctaviaException()]
        update_json = {'description': 'Some availability zone.'}
        body = self._build_body(update_json)
        response = self.put(self.AZ_PATH.format(az_name='bogus'), body,
                            status=500)
        self.assertEqual('An unknown exception occurred.',
                         response.json.get('faultstring'))

    def test_delete(self):
        az = self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        self.delete(self.AZ_PATH.format(az_name=az.get('name')))
        response = self.get(self.AZ_PATH.format(az_name=az.get('name')),
                            status=404)
        err_msg = "Availability Zone %s not found." % az.get('name')
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_delete_nonexistent_name(self):
        response = self.delete(
            self.AZ_PATH.format(az_name='bogus_name'), status=404)
        self.assertEqual('Availability Zone bogus_name not found.',
                         response.json.get('faultstring'))

    def test_delete_deleted_name(self):
        response = self.delete(
            self.AZ_PATH.format(az_name=constants.NIL_UUID), status=404)
        self.assertEqual(
            'Availability Zone {} not found.'.format(constants.NIL_UUID),
            response.json.get('faultstring'))

    def test_delete_authorized(self):
        az = self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                self.delete(
                    self.AZ_PATH.format(az_name=az.get('name')))
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        response = self.get(self.AZ_PATH.format(az_name=az.get('name')),
                            status=404)
        err_msg = "Availability Zone %s not found." % az.get('name')
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_delete_not_authorized(self):
        az = self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)

        response = self.delete(self.AZ_PATH.format(az_name=az.get('name')),
                               status=403)
        api_az = response.json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, api_az)

        response = self.get(self.AZ_PATH.format(
            az_name=az.get('name'))).json.get(self.root_tag)
        self.assertEqual('name1', response.get('name'))

    def test_delete_in_use(self):
        az = self.create_availability_zone(
            'name1', 'description', self.azp.get('id'), True)
        project_id = uuidutils.generate_uuid()
        lb_id = uuidutils.generate_uuid()
        self.create_load_balancer(lb_id, name='lb1',
                                  project_id=project_id,
                                  description='desc1',
                                  availability_zone=az.get('name'),
                                  admin_state_up=False)
        self.delete(self.AZ_PATH.format(az_name=az.get('name')),
                    status=409)
        response = self.get(self.AZ_PATH.format(
            az_name=az.get('name'))).json.get(self.root_tag)
        self.assertEqual('name1', response.get('name'))

    @mock.patch('octavia.db.repositories.AvailabilityZoneRepository.delete')
    def test_delete_exception(self, mock_delete):
        mock_delete.side_effect = [exceptions.OctaviaException()]
        response = self.delete(self.AZ_PATH.format(az_name='bogus'),
                               status=500)
        self.assertEqual('An unknown exception occurred.',
                         response.json.get('faultstring'))



#    Copyright 2017 GoDaddy
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""amphora add image id

Revision ID: 034756a182a2
Revises: 10d38216ad34
Create Date: 2018-02-26 17:38:37.971677

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '034756a182a2'
down_revision = '10d38216ad34'


def upgrade():
    op.add_column(
        u'amphora',
        sa.Column(u'image_id', sa.String(36), nullable=True)
    )



# Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import string

from novaclient import exceptions as nova_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from stevedore import driver as stevedore_driver

from octavia.common import clients
from octavia.common import constants
from octavia.common import data_models as models
from octavia.common import exceptions
from octavia.compute import compute_base

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


def _extract_amp_image_id_by_tag(client, image_tag, image_owner):
    if image_owner:
        images = list(client.images.list(
            filters={'tag': [image_tag],
                     'owner': image_owner,
                     'status': constants.GLANCE_IMAGE_ACTIVE},
            sort='created_at:desc',
            limit=2))
    else:
        images = list(client.images.list(
            filters={'tag': [image_tag],
                     'status': constants.GLANCE_IMAGE_ACTIVE},
            sort='created_at:desc',
            limit=2))

    if not images:
        raise exceptions.GlanceNoTaggedImages(tag=image_tag)
    image_id = images[0]['id']
    num_images = len(images)
    if num_images > 1:
        LOG.warning("A single Glance image should be tagged with %(tag)s tag, "
                    "but at least two were found. Using %(image_id)s.",
                    {'tag': image_tag, 'image_id': image_id})
    return image_id


def _get_image_uuid(client, image_id, image_tag, image_owner):
    if image_id:
        if image_tag:
            LOG.warning("Both amp_image_id and amp_image_tag options defined. "
                        "Using the amp_image_id.")
        return image_id

    return _extract_amp_image_id_by_tag(client, image_tag, image_owner)


class VirtualMachineManager(compute_base.ComputeBase):
    '''Compute implementation of virtual machines via nova.'''

    def __init__(self):
        super(VirtualMachineManager, self).__init__()
        # Must initialize nova api
        self._nova_client = clients.NovaAuth.get_nova_client(
            endpoint=CONF.nova.endpoint,
            region=CONF.nova.region_name,
            endpoint_type=CONF.nova.endpoint_type,
            insecure=CONF.nova.insecure,
            cacert=CONF.nova.ca_certificates_file)
        self._glance_client = clients.GlanceAuth.get_glance_client(
            service_name=CONF.glance.service_name,
            endpoint=CONF.glance.endpoint,
            region=CONF.glance.region_name,
            endpoint_type=CONF.glance.endpoint_type,
            insecure=CONF.glance.insecure,
            cacert=CONF.glance.ca_certificates_file)
        self.manager = self._nova_client.servers
        self.server_groups = self._nova_client.server_groups
        self.flavor_manager = self._nova_client.flavors
        self.availability_zone_manager = self._nova_client.availability_zones
        self.volume_driver = stevedore_driver.DriverManager(
            namespace='octavia.volume.drivers',
            name=CONF.controller_worker.volume_driver,
            invoke_on_load=True
        ).driver

    def build(self, name="amphora_name", amphora_flavor=None,
              image_id=None, image_tag=None, image_owner=None,
              key_name=None, sec_groups=None, network_ids=None,
              port_ids=None, config_drive_files=None, user_data=None,
              server_group_id=None, availability_zone=None):
        '''Create a new virtual machine.

        :param name: optional name for amphora
        :param amphora_flavor: image flavor for virtual machine
        :param image_id: image ID for virtual machine
        :param image_tag: image tag for virtual machine
        :param key_name: keypair to add to the virtual machine
        :param sec_groups: Security group IDs for virtual machine
        :param network_ids: Network IDs to include on virtual machine
        :param port_ids: Port IDs to include on virtual machine
        :param config_drive_files:  An optional dict of files to overwrite on
                                    the server upon boot. Keys are file names
                                    (i.e. /etc/passwd) and values are the file
                                    contents (either as a string or as a
                                    file-like object). A maximum of five
                                    entries is allowed, and each file must be
                                    10k or less.
        :param user_data: Optional user data to pass to be exposed by the
                          metadata server this can be a file type object as
                          well or a string
        :param server_group_id: Optional server group id(uuid) which is used
                                for anti_affinity feature
        :param availability_zone: Name of the compute availability zone.

        :raises ComputeBuildException: if nova failed to build virtual machine
        :returns: UUID of amphora

        '''

        volume_id = None
        try:
            network_ids = network_ids or []
            port_ids = port_ids or []
            nics = []
            if network_ids:
                nics.extend([{"net-id": net_id} for net_id in network_ids])
            if port_ids:
                nics.extend([{"port-id": port_id} for port_id in port_ids])

            server_group = None if server_group_id is None else {
                "group": server_group_id}
            az_name = availability_zone or CONF.nova.availability_zone

            image_id = _get_image_uuid(
                self._glance_client, image_id, image_tag, image_owner)

            if CONF.nova.random_amphora_name_length:
                r = random.SystemRandom()
                name = "a{}".format("".join(
                    [r.choice(string.ascii_uppercase + string.digits)
                     for i in range(CONF.nova.random_amphora_name_length - 1)]
                ))
            block_device_mapping = {}
            if (CONF.controller_worker.volume_driver !=
                    constants.VOLUME_NOOP_DRIVER):
                # creating volume
                LOG.debug('Creating volume for amphora from image %s',
                          image_id)
                volume_id = self.volume_driver.create_volume_from_image(
                    image_id)
                LOG.debug('Created boot volume %s for amphora', volume_id)
                # If use volume based, does not require image ID anymore
                image_id = None
                # Boot from volume with parameters: target device name = vda,
                # device id = volume_id, device type and size unspecified,
                # delete-on-terminate = true (volume will be deleted by Nova
                # on instance termination)
                block_device_mapping = {'vda': '%s:::true' % volume_id}
            amphora = self.manager.create(
                name=name, image=image_id, flavor=amphora_flavor,
                block_device_mapping=block_device_mapping,
                key_name=key_name, security_groups=sec_groups,
                nics=nics,
                files=config_drive_files,
                userdata=user_data,
                config_drive=True,
                scheduler_hints=server_group,
                availability_zone=az_name
            )

            return amphora.id
        except Exception as e:
            if (CONF.controller_worker.volume_driver !=
                    constants.VOLUME_NOOP_DRIVER):
                self.volume_driver.delete_volume(volume_id)
            LOG.exception("Nova failed to build the instance due to: %s",
                          str(e))
            raise exceptions.ComputeBuildException(fault=e)

    def delete(self, compute_id):
        '''Delete a virtual machine.

        :param compute_id: virtual machine UUID
        '''
        try:
            self.manager.delete(server=compute_id)
        except nova_exceptions.NotFound:
            LOG.warning("Nova instance with id: %s not found. "
                        "Assuming already deleted.", compute_id)
        except Exception as e:
            LOG.exception("Error deleting nova virtual machine.")
            raise exceptions.ComputeDeleteException(compute_msg=str(e))

    def status(self, compute_id):
        '''Retrieve the status of a virtual machine.

        :param compute_id: virtual machine UUID
        :returns: constant of amphora status
        '''
        try:
            amphora, fault = self.get_amphora(compute_id)
            if amphora and amphora.status == 'ACTIVE':
                return constants.UP
        except Exception:
            LOG.exception("Error retrieving nova virtual machine status.")
            raise exceptions.ComputeStatusException()
        return constants.DOWN

    def get_amphora(self, compute_id, management_network_id=None):
        '''Retrieve the information in nova of a virtual machine.

        :param compute_id: virtual machine UUID
        :param management_network_id: ID of the management network
        :returns: an amphora object
        :returns: fault message or None
        '''
        # utilize nova client ServerManager 'get' method to retrieve info
        try:
            amphora = self.manager.get(compute_id)
        except Exception:
            LOG.exception("Error retrieving nova virtual machine.")
            raise exceptions.ComputeGetException()
        return self._translate_amphora(amphora, management_network_id)

    def _translate_amphora(self, nova_response, management_network_id=None):
        '''Convert a nova virtual machine into an amphora object.

        :param nova_response: JSON response from nova
        :param management_network_id: ID of the management network
        :returns: an amphora object
        :returns: fault message or None
        '''
        # Extract interfaces of virtual machine to populate desired amphora
        # fields

        lb_network_ip = None
        availability_zone = None
        image_id = None

        if management_network_id:
            boot_networks = [management_network_id]
        else:
            boot_networks = CONF.controller_worker.amp_boot_network_list

        try:
            inf_list = nova_response.interface_list()
            for interface in inf_list:
                net_id = interface.net_id
                # Pick the first fixed_ip if this is a boot network or if
                # there are no boot networks configured (use default network)
                if net_id in boot_networks or not boot_networks:
                    lb_network_ip = interface.fixed_ips[0]['ip_address']
                    break
            try:
                availability_zone = getattr(
                    nova_response, 'OS-EXT-AZ:availability_zone')
            except AttributeError:
                LOG.info('No availability zone listed for server %s',
                         nova_response.id)
        except Exception:
            LOG.debug('Extracting virtual interfaces through nova '
                      'os-interfaces extension failed.')

        fault = getattr(nova_response, 'fault', None)
        if (CONF.controller_worker.volume_driver ==
                constants.VOLUME_NOOP_DRIVER):
            image_id = nova_response.image.get("id")
        else:
            try:
                volumes = self._nova_client.volumes.get_server_volumes(
                    nova_response.id)
            except Exception:
                LOG.debug('Extracting volumes through nova '
                          'os-volumes extension failed.')
                volumes = []
            if not volumes:
                LOG.warning('Boot volume not found for volume backed '
                            'amphora instance %s ', nova_response.id)
            else:
                if len(volumes) > 1:
                    LOG.warning('Found more than one (%s) volumes '
                                'for amphora instance %s',
                                len(volumes), nova_response.id)
                volume_id = volumes[0].volumeId
                image_id = self.volume_driver.get_image_from_volume(volume_id)

        response = models.Amphora(
            compute_id=nova_response.id,
            status=nova_response.status,
            lb_network_ip=lb_network_ip,
            cached_zone=availability_zone,
            image_id=image_id,
            compute_flavor=nova_response.flavor.get("id")
        )
        return response, fault

    def create_server_group(self, name, policy):
        """Create a server group object

        :param name: the name of the server group
        :param policy: the policy of the server group
        :raises: Generic exception if the server group is not created
        :returns: the server group object
        """
        kwargs = {'name': name,
                  'policies': [policy]}
        try:
            server_group_obj = self.server_groups.create(**kwargs)
            return server_group_obj
        except Exception:
            LOG.exception("Error create server group instance.")
            raise exceptions.ServerGroupObjectCreateException()

    def delete_server_group(self, server_group_id):
        """Delete a server group object

        :raises: Generic exception if the server group is not deleted
        :param server_group_id: the uuid of a server group
        """
        try:
            self.server_groups.delete(server_group_id)

        except nova_exceptions.NotFound:
            LOG.warning("Server group instance with id: %s not found. "
                        "Assuming already deleted.", server_group_id)
        except Exception:
            LOG.exception("Error delete server group instance.")
            raise exceptions.ServerGroupObjectDeleteException()

    def attach_network_or_port(self, compute_id, network_id=None,
                               ip_address=None, port_id=None):
        """Attaching a port or a network to an existing amphora

        :param compute_id: id of an amphora in the compute service
        :param network_id: id of a network
        :param ip_address: ip address to attempt to be assigned to interface
        :param port_id: id of the neutron port
        :return: nova interface instance
        :raises ComputePortInUseException: The port is in use somewhere else
        :raises ComputeUnknownException: Unknown nova error
        """
        try:
            interface = self.manager.interface_attach(
                server=compute_id, net_id=network_id, fixed_ip=ip_address,
                port_id=port_id)
        except nova_exceptions.Conflict as e:
            # The port is already in use.
            if port_id:
                # Check if the port we want is already attached
                try:
                    interfaces = self.manager.interface_list(compute_id)
                    for interface in interfaces:
                        if interface.id == port_id:
                            return interface
                except Exception as e:
                    raise exceptions.ComputeUnknownException(exc=str(e))

                raise exceptions.ComputePortInUseException(port=port_id)

            # Nova should have created the port, so something is really
            # wrong in nova if we get here.
            raise exceptions.ComputeUnknownException(exc=str(e))
        except nova_exceptions.NotFound as e:
            if 'Instance' in str(e):
                raise exceptions.NotFound(resource='Instance', id=compute_id)
            if 'Network' in str(e):
                raise exceptions.NotFound(resource='Network', id=network_id)
            if 'Port' in str(e):
                raise exceptions.NotFound(resource='Port', id=port_id)
            raise exceptions.NotFound(resource=str(e), id=compute_id)
        except Exception as e:
            LOG.error('Error attaching network %(network_id)s with ip '
                      '%(ip_address)s and port %(port)s to amphora '
                      '(compute_id: %(compute_id)s) ',
                      {
                          'compute_id': compute_id,
                          'network_id': network_id,
                          'ip_address': ip_address,
                          'port': port_id
                      })
            raise exceptions.ComputeUnknownException(exc=str(e))
        return interface

    def detach_port(self, compute_id, port_id):
        """Detaches a port from an existing amphora.

        :param compute_id: id of an amphora in the compute service
        :param port_id: id of the port
        :return: None
        """
        try:
            self.manager.interface_detach(server=compute_id,
                                          port_id=port_id)
        except Exception:
            LOG.error('Error detaching port %(port_id)s from amphora '
                      'with compute ID %(compute_id)s. '
                      'Skipping.',
                      {
                          'port_id': port_id,
                          'compute_id': compute_id
                      })

    def validate_flavor(self, flavor_id):
        """Validates that a flavor exists in nova.

        :param flavor_id: ID of the flavor to lookup in nova.
        :raises: NotFound
        :returns: None
        """
        try:
            self.flavor_manager.get(flavor_id)
        except nova_exceptions.NotFound:
            LOG.info('Flavor %s was not found in nova.', flavor_id)
            raise exceptions.InvalidSubresource(resource='Nova flavor',
                                                id=flavor_id)
        except Exception as e:
            LOG.exception('Nova reports a failure getting flavor details for '
                          'flavor ID %s: %s', flavor_id, str(e))
            raise

    def validate_availability_zone(self, availability_zone):
        """Validates that an availability zone exists in nova.

        :param availability_zone: Name of the availability zone to lookup.
        :raises: NotFound
        :returns: None
        """
        try:
            compute_zones = [
                a.zoneName for a in self.availability_zone_manager.list(
                    detailed=False)]
            if availability_zone not in compute_zones:
                LOG.info('Availability zone %s was not found in nova. %s',
                         availability_zone, compute_zones)
                raise exceptions.InvalidSubresource(
                    resource='Nova availability zone', id=availability_zone)
        except Exception as e:
            LOG.exception('Nova reports a failure getting listing '
                          'availability zones: %s', str(e))
            raise



#    Copyright 2015 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""L7 Policies and Rules

Revision ID: 458c9ee2a011
Revises: 29ff921a6eb
Create Date: 2016-01-07 11:45:45.391851

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '458c9ee2a011'
down_revision = '29ff921a6eb'


def upgrade():
    # L7 Rule Types
    op.create_table(
        u'l7rule_type',
        sa.Column(u'name', sa.String(36), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    # Create temporary table for table data seeding
    insert_table = sql.table(
        u'l7rule_type',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'HOST_NAME'},
            {'name': 'PATH'},
            {'name': 'FILE_TYPE'},
            {'name': 'HEADER'},
            {'name': 'COOKIE'}
        ]
    )

    # L7 Rule Compare Types
    op.create_table(
        u'l7rule_compare_type',
        sa.Column(u'name', sa.String(36), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'l7rule_compare_type',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'REGEX'},
            {'name': 'STARTS_WITH'},
            {'name': 'ENDS_WITH'},
            {'name': 'CONTAINS'},
            {'name': 'EQUAL_TO'}
        ]
    )

    # L7 Policy Actions
    op.create_table(
        u'l7policy_action',
        sa.Column(u'name', sa.String(36), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'l7policy_action',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'REJECT'},
            {'name': 'REDIRECT_TO_URL'},
            {'name': 'REDIRECT_TO_POOL'}
        ]
    )

    # L7 Policies
    op.create_table(
        u'l7policy',
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'listener_id', sa.String(36), nullable=False),
        sa.Column(u'action', sa.String(36), nullable=False),
        sa.Column(u'redirect_pool_id', sa.String(36), nullable=True),
        sa.Column(u'redirect_url', sa.String(255), nullable=True),
        sa.Column(u'position', sa.Integer, nullable=False),
        sa.Column(u'enabled', sa.Boolean(), default=True, nullable=False),

        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'listener_id'],
                                [u'listener.id'],
                                name=u'fk_l7policy_listener_id'),
        sa.ForeignKeyConstraint([u'redirect_pool_id'],
                                [u'pool.id'],
                                name=u'fk_l7policy_pool_id'),
        sa.ForeignKeyConstraint([u'action'],
                                [u'l7policy_action.name'],
                                name=u'fk_l7policy_l7policy_action_name')
    )

    # L7 Rules
    op.create_table(
        u'l7rule',
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'l7policy_id', sa.String(36), nullable=False),
        sa.Column(u'type', sa.String(36), nullable=False),
        sa.Column(u'compare_type', sa.String(36), nullable=False),
        sa.Column(u'key', sa.String(255), nullable=True),
        sa.Column(u'value', sa.String(255), nullable=False),
        sa.Column(u'invert', sa.Boolean(), default=False, nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'l7policy_id'],
                                [u'l7policy.id'],
                                name=u'fk_l7rule_l7policy_id'),
        sa.ForeignKeyConstraint([u'type'],
                                [u'l7rule_type.name'],
                                name=u'fk_l7rule_l7rule_type_name'),
        sa.ForeignKeyConstraint([u'compare_type'],
                                [u'l7rule_compare_type.name'],
                                name=u'fk_l7rule_l7rule_compare_type_name')
    )



# Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from barbicanclient.v1 import containers
from barbicanclient.v1 import secrets

import octavia.certificates.common.barbican as barbican_common
from octavia.common import utils as octavia_utils
import octavia.tests.common.sample_certs as sample
import octavia.tests.unit.base as base


class TestBarbicanCert(base.TestCase):

    def _prepare(self):
        self.certificate_secret = secrets.Secret(
            api=mock.MagicMock(),
            payload=self.certificate
        )
        self.intermediates_secret = secrets.Secret(
            api=mock.MagicMock(),
            payload=sample.X509_IMDS
        )
        self.private_key_secret = secrets.Secret(
            api=mock.MagicMock(),
            payload=self.private_key
        )
        self.private_key_passphrase_secret = secrets.Secret(
            api=mock.MagicMock(),
            payload=self.private_key_passphrase
        )

    def test_barbican_cert(self):
        # Certificate data
        self.certificate = bytes(sample.X509_CERT)
        self.intermediates = sample.X509_IMDS_LIST
        self.private_key = bytes(sample.X509_CERT_KEY_ENCRYPTED)
        self.private_key_passphrase = sample.X509_CERT_KEY_PASSPHRASE
        self._prepare()

        container = containers.CertificateContainer(
            api=mock.MagicMock(),
            certificate=self.certificate_secret,
            intermediates=self.intermediates_secret,
            private_key=self.private_key_secret,
            private_key_passphrase=self.private_key_passphrase_secret
        )
        # Create a cert
        cert = barbican_common.BarbicanCert(
            cert_container=container
        )

        # Validate the cert functions
        self.assertEqual(cert.get_certificate(), sample.X509_CERT)
        self.assertEqual(cert.get_intermediates(), sample.X509_IMDS_LIST)
        self.assertEqual(cert.get_private_key(),
                         sample.X509_CERT_KEY_ENCRYPTED)
        self.assertEqual(cert.get_private_key_passphrase(),
                         octavia_utils.b(sample.X509_CERT_KEY_PASSPHRASE))

    def test_barbican_cert_text(self):
        # Certificate data
        self.certificate = str(sample.X509_CERT)
        self.intermediates = str(sample.X509_IMDS_LIST)
        self.private_key = str(sample.X509_CERT_KEY_ENCRYPTED)
        self.private_key_passphrase = str(sample.X509_CERT_KEY_PASSPHRASE)
        self._prepare()

        container = containers.CertificateContainer(
            api=mock.MagicMock(),
            certificate=self.certificate_secret,
            intermediates=self.intermediates_secret,
            private_key=self.private_key_secret,
            private_key_passphrase=self.private_key_passphrase_secret
        )
        # Create a cert
        cert = barbican_common.BarbicanCert(
            cert_container=container
        )

        # Validate the cert functions
        self.assertEqual(cert.get_certificate(),
                         octavia_utils.b(str(sample.X509_CERT)))
        self.assertEqual(cert.get_intermediates(), sample.X509_IMDS_LIST)
        self.assertEqual(cert.get_private_key(), octavia_utils.b(str(
            sample.X509_CERT_KEY_ENCRYPTED)))
        self.assertEqual(cert.get_private_key_passphrase(),
                         octavia_utils.b(sample.X509_CERT_KEY_PASSPHRASE))



#    Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import errno
from unittest import mock

from oslo_config import cfg
from oslo_serialization import jsonutils

from octavia.api.drivers.driver_agent import driver_listener
import octavia.tests.unit.base as base

CONF = cfg.CONF


class TestDriverListener(base.TestCase):

    def setUp(self):
        super(TestDriverListener, self).setUp()

    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.memoryview')
    def test_recv(self, mock_memoryview):
        # TEST_STRING len() is 15
        TEST_STRING = '{"test": "msg"}'
        ref_object = jsonutils.loads(TEST_STRING)

        mock_recv_socket = mock.MagicMock()
        mock_recv = mock.MagicMock()
        mock_recv.side_effect = [b'1', b'5', b'\n']
        mock_recv_socket.recv = mock_recv
        mock_recv_socket.recv_into.return_value = 15
        mock_mv_buffer = mock.MagicMock()
        mock_tobytes = mock.MagicMock()
        mock_tobytes.return_value = TEST_STRING
        mock_mv_buffer.tobytes = mock_tobytes
        mock_memoryview.return_value = mock_mv_buffer

        result = driver_listener._recv(mock_recv_socket)

        self.assertEqual(ref_object, result)
        calls = [mock.call(1), mock.call(1), mock.call(1)]
        mock_recv.assert_has_calls(calls)
        mock_memoryview.assert_called_once_with(bytearray(15))
        mock_recv_socket.recv_into.assert_called_once_with(mock_mv_buffer[0:],
                                                           15)

    @mock.patch('octavia.api.drivers.driver_agent.driver_updater.'
                'DriverUpdater')
    @mock.patch('octavia.api.drivers.driver_agent.driver_listener._recv')
    def test_StatusRequestHandler_handle(self, mock_recv, mock_driverupdater):
        TEST_OBJECT = {"test": "msg"}
        mock_recv.return_value = 'bogus'
        mock_updater = mock.MagicMock()
        mock_update_loadbalancer_status = mock.MagicMock()
        mock_update_loadbalancer_status.return_value = TEST_OBJECT
        mock_updater.update_loadbalancer_status = (
            mock_update_loadbalancer_status)
        mock_driverupdater.return_value = mock_updater
        mock_request = mock.MagicMock()
        mock_send = mock.MagicMock()
        mock_sendall = mock.MagicMock()
        mock_request.send = mock_send
        mock_request.sendall = mock_sendall

        StatusRequestHandler = driver_listener.StatusRequestHandler(
            mock_request, 'bogus', 'bogus')
        StatusRequestHandler.handle()

        mock_recv.assert_called_with(mock_request)
        mock_update_loadbalancer_status.assert_called_with('bogus')
        mock_send.assert_called_with(b'15\n')
        mock_sendall.assert_called_with(
            jsonutils.dumps(TEST_OBJECT).encode('utf-8'))

    @mock.patch('octavia.api.drivers.driver_agent.driver_updater.'
                'DriverUpdater')
    @mock.patch('octavia.api.drivers.driver_agent.driver_listener._recv')
    def test_StatsRequestHandler_handle(self, mock_recv, mock_driverupdater):
        TEST_OBJECT = {"test": "msg"}
        mock_recv.return_value = 'bogus'
        mock_updater = mock.MagicMock()
        mock_update_listener_stats = mock.MagicMock()
        mock_update_listener_stats.return_value = TEST_OBJECT
        mock_updater.update_listener_statistics = (mock_update_listener_stats)
        mock_driverupdater.return_value = mock_updater
        mock_request = mock.MagicMock()
        mock_send = mock.MagicMock()
        mock_sendall = mock.MagicMock()
        mock_request.send = mock_send
        mock_request.sendall = mock_sendall

        StatsRequestHandler = driver_listener.StatsRequestHandler(
            mock_request, 'bogus', 'bogus')
        StatsRequestHandler.handle()

        mock_recv.assert_called_with(mock_request)
        mock_update_listener_stats.assert_called_with('bogus')
        mock_send.assert_called_with(b'15\n')
        mock_sendall.assert_called_with(jsonutils.dump_as_bytes(TEST_OBJECT))

    @mock.patch('octavia.api.drivers.driver_agent.driver_get.'
                'process_get')
    @mock.patch('octavia.api.drivers.driver_agent.driver_listener._recv')
    def test_GetRequestHandler_handle(self, mock_recv, mock_process_get):
        TEST_OBJECT = {"test": "msg"}

        mock_recv.return_value = 'bogus'

        mock_process_get.return_value = TEST_OBJECT
        mock_request = mock.MagicMock()
        mock_send = mock.MagicMock()
        mock_sendall = mock.MagicMock()
        mock_request.send = mock_send
        mock_request.sendall = mock_sendall

        GetRequestHandler = driver_listener.GetRequestHandler(
            mock_request, 'bogus', 'bogus')
        GetRequestHandler.handle()

        mock_recv.assert_called_with(mock_request)
        mock_process_get.assert_called_with('bogus')

        mock_send.assert_called_with(b'15\n')
        mock_sendall.assert_called_with(jsonutils.dump_as_bytes(TEST_OBJECT))

    @mock.patch('os.remove')
    def test_cleanup_socket_file(self, mock_remove):
        mock_remove.side_effect = [mock.DEFAULT, OSError,
                                   OSError(errno.ENOENT, 'no_file')]
        driver_listener._cleanup_socket_file('fake_filename')
        mock_remove.assert_called_once_with('fake_filename')

        self.assertRaises(OSError, driver_listener._cleanup_socket_file,
                          'fake_filename')
        # Make sure we just pass if the file was not found
        driver_listener._cleanup_socket_file('fake_filename')

    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.'
                '_cleanup_socket_file')
    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.'
                'ForkingUDSServer')
    def test_status_listener(self, mock_forking_server, mock_cleanup):
        mock_server = mock.MagicMock()
        mock_active_children = mock.PropertyMock(
            side_effect=['a', 'a', 'a',
                         'a' * CONF.driver_agent.status_max_processes, 'a',
                         'a' * 1000, ''])
        type(mock_server).active_children = mock_active_children
        mock_forking_server.return_value = mock_server
        mock_exit_event = mock.MagicMock()
        mock_exit_event.is_set.side_effect = [False, False, False, False, True]

        driver_listener.status_listener(mock_exit_event)
        mock_server.handle_request.assert_called()
        mock_server.server_close.assert_called_once()
        self.assertEqual(2, mock_cleanup.call_count)

    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.'
                '_cleanup_socket_file')
    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.'
                'ForkingUDSServer')
    def test_stats_listener(self, mock_forking_server, mock_cleanup):
        mock_server = mock.MagicMock()
        mock_active_children = mock.PropertyMock(
            side_effect=['a', 'a', 'a',
                         'a' * CONF.driver_agent.status_max_processes, 'a',
                         'a' * 1000, ''])
        type(mock_server).active_children = mock_active_children
        mock_forking_server.return_value = mock_server
        mock_exit_event = mock.MagicMock()
        mock_exit_event.is_set.side_effect = [False, False, False, False, True]

        driver_listener.stats_listener(mock_exit_event)
        mock_server.handle_request.assert_called()
        mock_server.server_close.assert_called_once()

    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.'
                '_cleanup_socket_file')
    @mock.patch('octavia.api.drivers.driver_agent.driver_listener.'
                'ForkingUDSServer')
    def test_get_listener(self, mock_forking_server, mock_cleanup):
        mock_server = mock.MagicMock()
        mock_active_children = mock.PropertyMock(
            side_effect=['a', 'a', 'a',
                         'a' * CONF.driver_agent.status_max_processes, 'a',
                         'a' * 1000, ''])
        type(mock_server).active_children = mock_active_children
        mock_forking_server.return_value = mock_server
        mock_exit_event = mock.MagicMock()
        mock_exit_event.is_set.side_effect = [False, False, False, False, True]

        driver_listener.get_listener(mock_exit_event)
        mock_server.handle_request.assert_called()
        mock_server.server_close.assert_called_once()



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import random
from unittest import mock

from octavia.amphorae.backends.utils import network_namespace
from octavia.tests.common import utils as test_utils
import octavia.tests.unit.base as base


class TestNetworkNamespace(base.TestCase):

    def setUp(self):
        super(TestNetworkNamespace, self).setUp()

    @mock.patch('ctypes.get_errno')
    @mock.patch('ctypes.CDLL')
    def test_error_handler(self, mock_cdll, mock_get_errno):
        FAKE_NETNS = 'fake-netns'
        netns = network_namespace.NetworkNamespace(FAKE_NETNS)

        # Test result 0
        netns._error_handler(0, None, None)

        mock_get_errno.assert_not_called()

        # Test result -1
        mock_get_errno.reset_mock()

        self.assertRaises(OSError, netns._error_handler, -1, None, None)

        mock_get_errno.assert_called_once_with()

    @mock.patch('os.getpid')
    @mock.patch('ctypes.CDLL')
    def test_init(self, mock_cdll, mock_getpid):
        FAKE_NETNS = 'fake-netns'
        FAKE_PID = random.randrange(100000)
        mock_cdll_obj = mock.MagicMock()
        mock_cdll.return_value = mock_cdll_obj
        mock_getpid.return_value = FAKE_PID
        expected_current_netns = '/proc/{pid}/ns/net'.format(pid=FAKE_PID)
        expected_target_netns = '/var/run/netns/{netns}'.format(
            netns=FAKE_NETNS)

        netns = network_namespace.NetworkNamespace(FAKE_NETNS)

        self.assertEqual(expected_current_netns, netns.current_netns)
        self.assertEqual(expected_target_netns, netns.target_netns)
        self.assertEqual(mock_cdll_obj.setns, netns.set_netns)
        self.assertEqual(netns.set_netns.errcheck, netns._error_handler)

    @mock.patch('os.getpid')
    @mock.patch('ctypes.CDLL')
    def test_enter(self, mock_cdll, mock_getpid):
        CLONE_NEWNET = 0x40000000
        FAKE_NETNS = 'fake-netns'
        FAKE_PID = random.randrange(100000)
        current_netns_fd = random.randrange(100000)
        target_netns_fd = random.randrange(100000)
        mock_getpid.return_value = FAKE_PID
        mock_cdll_obj = mock.MagicMock()
        mock_cdll.return_value = mock_cdll_obj
        expected_current_netns = '/proc/{pid}/ns/net'.format(pid=FAKE_PID)
        expected_target_netns = '/var/run/netns/{netns}'.format(
            netns=FAKE_NETNS)

        netns = network_namespace.NetworkNamespace(FAKE_NETNS)

        current_mock_open = self.useFixture(
            test_utils.OpenFixture(expected_current_netns)).mock_open
        current_mock_open.return_value = current_netns_fd

        target_mock_open = self.useFixture(
            test_utils.OpenFixture(expected_target_netns)).mock_open
        handle = target_mock_open()
        handle.fileno.return_value = target_netns_fd

        netns.__enter__()

        self.assertEqual(current_netns_fd, netns.current_netns_fd)
        netns.set_netns.assert_called_once_with(target_netns_fd, CLONE_NEWNET)

    @mock.patch('os.getpid')
    @mock.patch('ctypes.CDLL')
    def test_exit(self, mock_cdll, mock_getpid):
        CLONE_NEWNET = 0x40000000
        FAKE_NETNS = 'fake-netns'
        FAKE_PID = random.randrange(100000)
        current_netns_fileno = random.randrange(100000)
        mock_getpid.return_value = FAKE_PID
        mock_cdll_obj = mock.MagicMock()
        mock_cdll.return_value = mock_cdll_obj
        mock_current_netns_fd = mock.MagicMock()
        mock_current_netns_fd.fileno.return_value = current_netns_fileno

        netns = network_namespace.NetworkNamespace(FAKE_NETNS)

        netns.current_netns_fd = mock_current_netns_fd

        netns.__exit__()

        netns.set_netns.assert_called_once_with(current_netns_fileno,
                                                CLONE_NEWNET)
        mock_current_netns_fd.close.assert_called_once_with()



# Copyright 2016 IBM
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import random
from unittest import mock

from oslo_utils import uuidutils

from octavia.common import constants
from octavia.common import data_models
from octavia.common import stats
from octavia.tests.unit import base


class TestStatsMixin(base.TestCase):

    def setUp(self):
        super(TestStatsMixin, self).setUp()
        self.sm = stats.StatsMixin()

        self.session = mock.MagicMock()
        self.listener_id = uuidutils.generate_uuid()
        self.amphora_id = uuidutils.generate_uuid()

        self.repo_listener_stats = mock.MagicMock()
        self.sm.listener_stats_repo = self.repo_listener_stats

        self.fake_stats = data_models.ListenerStatistics(
            listener_id=self.listener_id,
            amphora_id=self.amphora_id,
            bytes_in=random.randrange(1000000000),
            bytes_out=random.randrange(1000000000),
            active_connections=random.randrange(1000000000),
            total_connections=random.randrange(1000000000),
            request_errors=random.randrange(1000000000))

        self.sm.listener_stats_repo.get_all.return_value = ([self.fake_stats],
                                                            None)

        self.repo_amphora = mock.MagicMock()
        self.sm.repo_amphora = self.repo_amphora

    def test_get_listener_stats(self):
        fake_amp = mock.MagicMock()
        fake_amp.status = constants.AMPHORA_ALLOCATED
        self.sm.repo_amphora.get.return_value = fake_amp

        ls_stats = self.sm.get_listener_stats(
            self.session, self.listener_id)
        self.repo_listener_stats.get_all.assert_called_once_with(
            self.session, listener_id=self.listener_id)
        self.repo_amphora.get.assert_called_once_with(
            self.session, id=self.amphora_id)

        self.assertEqual(self.fake_stats.bytes_in, ls_stats.bytes_in)
        self.assertEqual(self.fake_stats.bytes_out, ls_stats.bytes_out)
        self.assertEqual(
            self.fake_stats.active_connections, ls_stats.active_connections)
        self.assertEqual(
            self.fake_stats.total_connections, ls_stats.total_connections)
        self.assertEqual(
            self.fake_stats.request_errors, ls_stats.request_errors)
        self.assertEqual(self.listener_id, ls_stats.listener_id)
        self.assertIsNone(ls_stats.amphora_id)

    def test_get_listener_stats_with_amphora_deleted(self):
        fake_amp = mock.MagicMock()
        fake_amp.status = constants.DELETED
        self.sm.repo_amphora.get.return_value = fake_amp

        ls_stats = self.sm.get_listener_stats(self.session, self.listener_id)
        self.repo_listener_stats.get_all.assert_called_once_with(
            self.session, listener_id=self.listener_id)
        self.repo_amphora.get.assert_called_once_with(
            self.session, id=self.amphora_id)

        self.assertEqual(self.fake_stats.bytes_in, ls_stats.bytes_in)
        self.assertEqual(self.fake_stats.bytes_out, ls_stats.bytes_out)
        self.assertEqual(0, ls_stats.active_connections)
        self.assertEqual(
            self.fake_stats.total_connections, ls_stats.total_connections)
        self.assertEqual(
            self.fake_stats.request_errors, ls_stats.request_errors)
        self.assertEqual(self.listener_id, ls_stats.listener_id)
        self.assertIsNone(ls_stats.amphora_id)



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_middleware import healthcheck
from pecan import abort as pecan_abort
from pecan import expose as pecan_expose
from pecan import request as pecan_request
from wsme import types as wtypes
from wsmeext import pecan as wsme_pecan

from octavia.api.v2 import controllers as v2_controller

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class RootController(object):
    """The controller with which the pecan wsgi app should be created."""

    def __init__(self):
        super(RootController, self).__init__()
        setattr(self, 'v2.0', v2_controller.V2Controller())
        setattr(self, 'v2', v2_controller.V2Controller())
        if CONF.api_settings.healthcheck_enabled:
            self.healthcheck_obj = healthcheck.Healthcheck.app_factory(None)

    # Run the oslo middleware healthcheck for /healthcheck
    @pecan_expose('json')
    @pecan_expose(content_type='plain/text')
    @pecan_expose(content_type='text/html')
    def healthcheck(self):  # pylint: disable=inconsistent-return-statements
        if CONF.api_settings.healthcheck_enabled:
            if pecan_request.method not in ['GET', 'HEAD']:
                pecan_abort(405)
            return self.healthcheck_obj.process_request(pecan_request)
        pecan_abort(404)

    def _add_a_version(self, versions, version, url_version, status,
                       timestamp, base_url):
        versions.append({
            'id': version,
            'status': status,
            'updated': timestamp,
            'links': [{
                'href': base_url + url_version,
                'rel': 'self'
            }]
        })

    @wsme_pecan.wsexpose(wtypes.text)
    def index(self):
        host_url = pecan_request.path_url

        if not host_url.endswith('/'):
            host_url = '{}/'.format(host_url)

        versions = []
        self._add_a_version(versions, 'v2.0', 'v2', 'SUPPORTED',
                            '2016-12-11T00:00:00Z', host_url)
        self._add_a_version(versions, 'v2.1', 'v2', 'SUPPORTED',
                            '2018-04-20T00:00:00Z', host_url)
        self._add_a_version(versions, 'v2.2', 'v2', 'SUPPORTED',
                            '2018-07-31T00:00:00Z', host_url)
        self._add_a_version(versions, 'v2.3', 'v2', 'SUPPORTED',
                            '2018-12-18T00:00:00Z', host_url)
        # amp statistics
        self._add_a_version(versions, 'v2.4', 'v2', 'SUPPORTED',
                            '2018-12-19T00:00:00Z', host_url)
        # Tags
        self._add_a_version(versions, 'v2.5', 'v2', 'SUPPORTED',
                            '2019-01-21T00:00:00Z', host_url)
        # Flavors
        self._add_a_version(versions, 'v2.6', 'v2', 'SUPPORTED',
                            '2019-01-25T00:00:00Z', host_url)
        # Amphora Config update
        self._add_a_version(versions, 'v2.7', 'v2', 'SUPPORTED',
                            '2018-01-25T12:00:00Z', host_url)
        # TLS client authentication
        self._add_a_version(versions, 'v2.8', 'v2', 'SUPPORTED',
                            '2019-02-12T00:00:00Z', host_url)
        # HTTP Redirect code
        self._add_a_version(versions, 'v2.9', 'v2', 'SUPPORTED',
                            '2019-03-04T00:00:00Z', host_url)
        # Healthmonitor host header
        self._add_a_version(versions, 'v2.10', 'v2', 'SUPPORTED',
                            '2019-03-05T00:00:00Z', host_url)
        # Additive batch member update
        self._add_a_version(versions, 'v2.11', 'v2', 'SUPPORTED',
                            '2019-06-24T00:00:00Z', host_url)
        # VIP ACL
        self._add_a_version(versions, 'v2.12', 'v2', 'SUPPORTED',
                            '2019-09-11T00:00:00Z', host_url)
        # SOURCE_IP_PORT algorithm
        self._add_a_version(versions, 'v2.13', 'v2', 'SUPPORTED',
                            '2019-09-13T00:00:00Z', host_url)
        # Availability Zones
        self._add_a_version(versions, 'v2.14', 'v2', 'SUPPORTED',
                            '2019-11-10T00:00:00Z', host_url)
        # TLS version and cipher options
        self._add_a_version(versions, 'v2.15', 'v2', 'SUPPORTED',
                            '2020-03-10T00:00:00Z', host_url)
        # Additional UDP Healthcheck Types (HTTP/TCP)
        self._add_a_version(versions, 'v2.16', 'v2', 'CURRENT',
                            '2020-03-15T00:00:00Z', host_url)
        return {'versions': versions}



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pecan import hooks

from octavia.api.common import pagination
from octavia.common import constants
from octavia.common import context


class ContextHook(hooks.PecanHook):
    """Configures a request context and attaches it to the request."""

    def on_route(self, state):
        context_obj = context.Context.from_environ(state.request.environ)
        state.request.context['octavia_context'] = context_obj


class QueryParametersHook(hooks.PecanHook):

    def before(self, state):
        if state.request.method != 'GET':
            return

        state.request.context[
            constants.PAGINATION_HELPER] = pagination.PaginationHelper(
            state.request.params.mixed())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Octavia."""
from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy as oslo_policy
from oslo_utils import excutils

from octavia.common import exceptions
from octavia import policies


LOG = logging.getLogger(__name__)
OCTAVIA_POLICY = None


def get_enforcer():
    global OCTAVIA_POLICY
    if OCTAVIA_POLICY is None:
        LOG.debug('Loading octavia policy object.')
        OCTAVIA_POLICY = Policy()
    return OCTAVIA_POLICY


def reset():
    global OCTAVIA_POLICY
    if OCTAVIA_POLICY:
        OCTAVIA_POLICY.clear()
        OCTAVIA_POLICY = None


class Policy(oslo_policy.Enforcer):

    def __init__(self, conf=cfg.CONF, policy_file=None, rules=None,
                 default_rule=None, use_conf=True, overwrite=True):
        """Init an Enforcer class.

            :param context: A context object.
            :param conf: A configuration object.
            :param policy_file: Custom policy file to use, if none is
                                specified, ``conf.oslo_policy.policy_file``
                                will be used.
            :param rules: Default dictionary / Rules to use. It will be
                          considered just in the first instantiation. If
                          :meth:`load_rules` with ``force_reload=True``,
                          :meth:`clear` or :meth:`set_rules` with
                          ``overwrite=True`` is called this will be
                          overwritten.
            :param default_rule: Default rule to use, conf.default_rule will
                                 be used if none is specified.
            :param use_conf: Whether to load rules from cache or config file.
            :param overwrite: Whether to overwrite existing rules when reload
                              rules from config file.
        """

        super(Policy, self).__init__(conf, policy_file, rules, default_rule,
                                     use_conf, overwrite)

        self.register_defaults(policies.list_rules())

    def authorize(self, action, target, context, do_raise=True, exc=None):
        """Verifies that the action is valid on the target in this context.

           :param context: The oslo context for this request.
           :param action: string representing the action to be checked
               this should be colon separated for clarity.
               i.e. ``compute:create_instance``,
               ``compute:attach_volume``,
               ``volume:attach_volume``
           :param target: dictionary representing the object of the action
               for object creation this should be a dictionary representing the
               location of the object e.g.
               ``{'project_id': context.project_id}``
           :param do_raise: if True (the default), raises PolicyForbidden;
               if False, returns False
           :param exc: Class of the exceptions to raise if the check fails.
                       Any remaining arguments passed to :meth:`enforce` (both
                       positional and keyword arguments) will be passed to
                       the exceptions class. If not specified,
                       :class:`PolicyForbidden` will be used.

           :raises PolicyForbidden: if verification fails
               and do_raise is True. Or if 'exc' is specified it will raise an
               exceptions of that type.

           :return: returns a non-False value (not necessarily "True") if
               authorized, and the exact value False if not authorized and
               do_raise is False.
        """
        credentials = context.to_policy_values()
        # Inject is_admin into the credentials to allow override via
        # config auth_strategy = constants.NOAUTH
        credentials['is_admin'] = (
            credentials.get('is_admin') or context.is_admin)

        if not exc:
            exc = exceptions.PolicyForbidden

        try:
            return super(Policy, self).authorize(
                action, target, credentials, do_raise=do_raise, exc=exc)
        except oslo_policy.PolicyNotRegistered:
            with excutils.save_and_reraise_exception():
                LOG.exception('Policy not registered')
        except Exception:
            credentials.pop('auth_token', None)
            with excutils.save_and_reraise_exception():
                LOG.debug('Policy check for %(action)s failed with '
                          'credentials %(credentials)s',
                          {'action': action, 'credentials': credentials})

    def check_is_admin(self, context):
        """Does roles contains 'admin' role according to policy setting.

        """
        credentials = context.to_dict()
        return self.enforce('context_is_admin', credentials, credentials)

    def get_rules(self):
        return self.rules


@oslo_policy.register('is_admin')
class IsAdminCheck(oslo_policy.Check):
    """An explicit check for is_admin."""

    def __init__(self, kind, match):
        """Initialize the check."""

        self.expected = match.lower() == 'true'

        super(IsAdminCheck, self).__init__(kind, str(self.expected))

    def __call__(self, target, creds, enforcer):
        """Determine whether is_admin matches the requested value."""

        return creds['is_admin'] == self.expected


# This is used for the oslopolicy-policy-generator tool
def get_no_context_enforcer():
    return Policy()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from oslo_config import cfg
from oslo_log import log as logging
from taskflow.patterns import linear_flow
from taskflow.patterns import unordered_flow

from octavia.common import constants
from octavia.common import exceptions
from octavia.common import utils
from octavia.controller.worker.v2.flows import amphora_flows
from octavia.controller.worker.v2.flows import listener_flows
from octavia.controller.worker.v2.flows import member_flows
from octavia.controller.worker.v2.flows import pool_flows
from octavia.controller.worker.v2.tasks import amphora_driver_tasks
from octavia.controller.worker.v2.tasks import compute_tasks
from octavia.controller.worker.v2.tasks import database_tasks
from octavia.controller.worker.v2.tasks import lifecycle_tasks
from octavia.controller.worker.v2.tasks import network_tasks
from octavia.db import repositories as repo

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class LoadBalancerFlows(object):

    def __init__(self):
        self.amp_flows = amphora_flows.AmphoraFlows()
        self.listener_flows = listener_flows.ListenerFlows()
        self.pool_flows = pool_flows.PoolFlows()
        self.member_flows = member_flows.MemberFlows()
        self.lb_repo = repo.LoadBalancerRepository()

    def get_create_load_balancer_flow(self, topology, listeners=None):
        """Creates a conditional graph flow that allocates a loadbalancer to

        two spare amphorae.
        :raises InvalidTopology: Invalid topology specified
        :return: The graph flow for creating a loadbalancer.
        """
        f_name = constants.CREATE_LOADBALANCER_FLOW
        lb_create_flow = linear_flow.Flow(f_name)

        lb_create_flow.add(lifecycle_tasks.LoadBalancerIDToErrorOnRevertTask(
            requires=constants.LOADBALANCER_ID))

        # allocate VIP
        lb_create_flow.add(database_tasks.ReloadLoadBalancer(
            name=constants.RELOAD_LB_BEFOR_ALLOCATE_VIP,
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER
        ))
        lb_create_flow.add(network_tasks.AllocateVIP(
            requires=constants.LOADBALANCER,
            provides=constants.VIP))
        lb_create_flow.add(database_tasks.UpdateVIPAfterAllocation(
            requires=(constants.LOADBALANCER_ID, constants.VIP),
            provides=constants.LOADBALANCER))
        lb_create_flow.add(network_tasks.UpdateVIPSecurityGroup(
            requires=constants.LOADBALANCER_ID))
        lb_create_flow.add(network_tasks.GetSubnetFromVIP(
            requires=constants.LOADBALANCER,
            provides=constants.SUBNET))

        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            lb_create_flow.add(*self._create_active_standby_topology())
        elif topology == constants.TOPOLOGY_SINGLE:
            lb_create_flow.add(*self._create_single_topology())
        else:
            LOG.error("Unknown topology: %s.  Unable to build load balancer.",
                      topology)
            raise exceptions.InvalidTopology(topology=topology)

        post_amp_prefix = constants.POST_LB_AMP_ASSOCIATION_SUBFLOW
        lb_create_flow.add(
            self.get_post_lb_amp_association_flow(
                post_amp_prefix, topology, mark_active=(not listeners)))

        if listeners:
            lb_create_flow.add(*self._create_listeners_flow())

        return lb_create_flow

    def _create_single_topology(self):
        sf_name = (constants.ROLE_STANDALONE + '-' +
                   constants.AMP_PLUG_NET_SUBFLOW)
        amp_for_lb_net_flow = linear_flow.Flow(sf_name)
        amp_for_lb_flow = self.amp_flows.get_amphora_for_lb_subflow(
            prefix=constants.ROLE_STANDALONE,
            role=constants.ROLE_STANDALONE)
        amp_for_lb_net_flow.add(amp_for_lb_flow)
        amp_for_lb_net_flow.add(*self._get_amp_net_subflow(sf_name))
        return amp_for_lb_net_flow

    def _create_active_standby_topology(
            self, lf_name=constants.CREATE_LOADBALANCER_FLOW):
        # When we boot up amphora for an active/standby topology,
        # we should leverage the Nova anti-affinity capabilities
        # to place the amphora on different hosts, also we need to check
        # if anti-affinity-flag is enabled or not:
        anti_affinity = CONF.nova.enable_anti_affinity
        flows = []
        if anti_affinity:
            # we need to create a server group first
            flows.append(
                compute_tasks.NovaServerGroupCreate(
                    name=lf_name + '-' +
                    constants.CREATE_SERVER_GROUP_FLOW,
                    requires=(constants.LOADBALANCER_ID),
                    provides=constants.SERVER_GROUP_ID))

            # update server group id in lb table
            flows.append(
                database_tasks.UpdateLBServerGroupInDB(
                    name=lf_name + '-' +
                    constants.UPDATE_LB_SERVERGROUPID_FLOW,
                    requires=(constants.LOADBALANCER_ID,
                              constants.SERVER_GROUP_ID)))

        f_name = constants.CREATE_LOADBALANCER_FLOW
        amps_flow = unordered_flow.Flow(f_name)

        master_sf_name = (constants.ROLE_MASTER + '-' +
                          constants.AMP_PLUG_NET_SUBFLOW)
        master_amp_sf = linear_flow.Flow(master_sf_name)
        master_amp_sf.add(self.amp_flows.get_amphora_for_lb_subflow(
            prefix=constants.ROLE_MASTER, role=constants.ROLE_MASTER))
        master_amp_sf.add(*self._get_amp_net_subflow(master_sf_name))

        backup_sf_name = (constants.ROLE_BACKUP + '-' +
                          constants.AMP_PLUG_NET_SUBFLOW)
        backup_amp_sf = linear_flow.Flow(backup_sf_name)
        backup_amp_sf.add(self.amp_flows.get_amphora_for_lb_subflow(
            prefix=constants.ROLE_BACKUP, role=constants.ROLE_BACKUP))
        backup_amp_sf.add(*self._get_amp_net_subflow(backup_sf_name))

        amps_flow.add(master_amp_sf, backup_amp_sf)

        return flows + [amps_flow]

    def _get_amp_net_subflow(self, sf_name):
        flows = []
        flows.append(network_tasks.PlugVIPAmphora(
            name=sf_name + '-' + constants.PLUG_VIP_AMPHORA,
            requires=(constants.LOADBALANCER, constants.AMPHORA,
                      constants.SUBNET),
            provides=constants.AMP_DATA))

        flows.append(network_tasks.ApplyQosAmphora(
            name=sf_name + '-' + constants.APPLY_QOS_AMP,
            requires=(constants.LOADBALANCER, constants.AMP_DATA,
                      constants.UPDATE_DICT)))
        flows.append(database_tasks.UpdateAmphoraVIPData(
            name=sf_name + '-' + constants.UPDATE_AMPHORA_VIP_DATA,
            requires=constants.AMP_DATA))
        flows.append(network_tasks.GetAmphoraNetworkConfigs(
            name=sf_name + '-' + constants.GET_AMP_NETWORK_CONFIG,
            requires=(constants.LOADBALANCER, constants.AMPHORA),
            provides=constants.AMPHORA_NETWORK_CONFIG))
        flows.append(amphora_driver_tasks.AmphoraPostVIPPlug(
            name=sf_name + '-' + constants.AMP_POST_VIP_PLUG,
            rebind={constants.AMPHORAE_NETWORK_CONFIG:
                    constants.AMPHORA_NETWORK_CONFIG},
            requires=(constants.LOADBALANCER,
                      constants.AMPHORAE_NETWORK_CONFIG)))
        return flows

    def _create_listeners_flow(self):
        flows = []
        flows.append(
            database_tasks.ReloadLoadBalancer(
                name=constants.RELOAD_LB_AFTER_AMP_ASSOC_FULL_GRAPH,
                requires=constants.LOADBALANCER_ID,
                provides=constants.LOADBALANCER
            )
        )
        flows.append(
            network_tasks.CalculateDelta(
                requires=(constants.LOADBALANCER, constants.AVAILABILITY_ZONE),
                provides=constants.DELTAS
            )
        )
        flows.append(
            network_tasks.HandleNetworkDeltas(
                requires=constants.DELTAS, provides=constants.ADDED_PORTS
            )
        )
        flows.append(
            amphora_driver_tasks.AmphoraePostNetworkPlug(
                requires=(constants.LOADBALANCER, constants.ADDED_PORTS)
            )
        )
        flows.append(
            self.listener_flows.get_create_all_listeners_flow()
        )
        flows.append(
            database_tasks.MarkLBActiveInDB(
                mark_subobjects=True,
                requires=constants.LOADBALANCER
            )
        )
        return flows

    def get_post_lb_amp_association_flow(self, prefix, topology,
                                         mark_active=True):
        """Reload the loadbalancer and create networking subflows for

        created/allocated amphorae.
        :return: Post amphorae association subflow
        """
        sf_name = prefix + '-' + constants.POST_LB_AMP_ASSOCIATION_SUBFLOW
        post_create_LB_flow = linear_flow.Flow(sf_name)
        post_create_LB_flow.add(
            database_tasks.ReloadLoadBalancer(
                name=sf_name + '-' + constants.RELOAD_LB_AFTER_AMP_ASSOC,
                requires=constants.LOADBALANCER_ID,
                provides=constants.LOADBALANCER))

        if topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            post_create_LB_flow.add(database_tasks.GetAmphoraeFromLoadbalancer(
                requires=constants.LOADBALANCER_ID,
                provides=constants.AMPHORAE))
            vrrp_subflow = self.amp_flows.get_vrrp_subflow(prefix)
            post_create_LB_flow.add(vrrp_subflow)

        post_create_LB_flow.add(database_tasks.UpdateLoadbalancerInDB(
            requires=[constants.LOADBALANCER, constants.UPDATE_DICT]))
        if mark_active:
            post_create_LB_flow.add(database_tasks.MarkLBActiveInDB(
                name=sf_name + '-' + constants.MARK_LB_ACTIVE_INDB,
                requires=constants.LOADBALANCER))
        return post_create_LB_flow

    def _get_delete_listeners_flow(self, listeners):
        """Sets up an internal delete flow

        :param listeners: A list of listener dicts
        :return: The flow for the deletion
        """
        listeners_delete_flow = unordered_flow.Flow('listeners_delete_flow')
        for listener in listeners:
            listeners_delete_flow.add(
                self.listener_flows.get_delete_listener_internal_flow(
                    listener))
        return listeners_delete_flow

    def get_delete_load_balancer_flow(self, lb):
        """Creates a flow to delete a load balancer.

        :returns: The flow for deleting a load balancer
        """
        return self._get_delete_load_balancer_flow(lb, False)

    def _get_delete_pools_flow(self, pools):
        """Sets up an internal delete flow

        Because task flow doesn't support loops we store each pool
        we want to delete in the store part and then rebind
        :param lb: load balancer
        :return: (flow, store) -- flow for the deletion and store with all
                    the listeners stored properly
        """
        pools_delete_flow = unordered_flow.Flow('pool_delete_flow')
        for pool in pools:
            pools_delete_flow.add(
                self.pool_flows.get_delete_pool_flow_internal(
                    pool[constants.POOL_ID]))
        return pools_delete_flow

    def _get_delete_load_balancer_flow(self, lb, cascade,
                                       listeners=(), pools=()):
        delete_LB_flow = linear_flow.Flow(constants.DELETE_LOADBALANCER_FLOW)
        delete_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(compute_tasks.NovaServerGroupDelete(
            requires=constants.SERVER_GROUP_ID))
        delete_LB_flow.add(database_tasks.MarkLBAmphoraeHealthBusy(
            requires=constants.LOADBALANCER))
        if cascade:
            listeners_delete = self._get_delete_listeners_flow(listeners)
            pools_delete = self._get_delete_pools_flow(pools)
            delete_LB_flow.add(pools_delete)
            delete_LB_flow.add(listeners_delete)
        delete_LB_flow.add(network_tasks.UnplugVIP(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(network_tasks.DeallocateVIP(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(compute_tasks.DeleteAmphoraeOnLoadBalancer(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.MarkLBAmphoraeDeletedInDB(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.DisableLBAmphoraeHealthMonitoring(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.MarkLBDeletedInDB(
            requires=constants.LOADBALANCER))
        delete_LB_flow.add(database_tasks.DecrementLoadBalancerQuota(
            requires=constants.PROJECT_ID))
        return delete_LB_flow

    def get_cascade_delete_load_balancer_flow(self, lb, listeners, pools):
        """Creates a flow to delete a load balancer.

        :returns: The flow for deleting a load balancer
        """
        return self._get_delete_load_balancer_flow(lb, True,
                                                   listeners=listeners,
                                                   pools=pools)

    def get_update_load_balancer_flow(self):
        """Creates a flow to update a load balancer.

        :returns: The flow for update a load balancer
        """
        update_LB_flow = linear_flow.Flow(constants.UPDATE_LOADBALANCER_FLOW)
        update_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))
        update_LB_flow.add(network_tasks.ApplyQos(
            requires=(constants.LOADBALANCER, constants.UPDATE_DICT)))
        update_LB_flow.add(amphora_driver_tasks.ListenersUpdate(
            requires=constants.LOADBALANCER_ID))
        update_LB_flow.add(database_tasks.UpdateLoadbalancerInDB(
            requires=[constants.LOADBALANCER, constants.UPDATE_DICT]))
        update_LB_flow.add(database_tasks.MarkLBActiveInDB(
            requires=constants.LOADBALANCER))

        return update_LB_flow

    def get_failover_LB_flow(self, amps, lb):
        """Failover a load balancer.

        1. Validate the VIP port is correct and present.
        2. Build a replacement amphora.
        3. Delete the failed amphora.
        4. Configure the replacement amphora listeners.
        5. Configure VRRP for the listeners.
        6. Build the second replacement amphora.
        7. Delete the second failed amphora.
        8. Delete any extraneous amphora.
        9. Configure the listeners on the new amphorae.
        10. Configure the VRRP on the new amphorae.
        11. Reload the listener configurations to pick up VRRP changes.
        12. Mark the load balancer back to ACTIVE.

        :returns: The flow that will provide the failover.
        """
        lb_topology = lb[constants.FLAVOR][constants.LOADBALANCER_TOPOLOGY]
        # Pick one amphora to be failed over if any exist.
        failed_amp = None
        if amps:
            failed_amp = amps.pop()

        failover_LB_flow = linear_flow.Flow(
            constants.FAILOVER_LOADBALANCER_FLOW)

        # Revert LB to provisioning_status ERROR if this flow goes wrong
        failover_LB_flow.add(lifecycle_tasks.LoadBalancerToErrorOnRevertTask(
            requires=constants.LOADBALANCER))

        # Setup timeouts for our requests to the amphorae
        timeout_dict = {
            constants.CONN_MAX_RETRIES:
                CONF.haproxy_amphora.active_connection_max_retries,
            constants.CONN_RETRY_INTERVAL:
                CONF.haproxy_amphora.active_connection_rety_interval}

        if failed_amp:
            failed_amp_role = failed_amp.get(constants.ROLE)
            if failed_amp_role in (constants.ROLE_MASTER,
                                   constants.ROLE_BACKUP):
                amp_role = 'master_or_backup'
            elif failed_amp_role == constants.ROLE_STANDALONE:
                amp_role = 'standalone'
            elif failed_amp_role is None:
                amp_role = 'spare'
            else:
                amp_role = 'undefined'
            LOG.info("Performing failover for amphora: %s",
                     {"id": failed_amp.get(constants.ID),
                      "load_balancer_id": lb.get(constants.ID),
                      "lb_network_ip": failed_amp.get(constants.LB_NETWORK_IP),
                      "compute_id": failed_amp.get(constants.COMPUTE_ID),
                      "role": amp_role})

            failover_LB_flow.add(database_tasks.MarkAmphoraPendingDeleteInDB(
                requires=constants.AMPHORA,
                inject={constants.AMPHORA: failed_amp}))

            failover_LB_flow.add(database_tasks.MarkAmphoraHealthBusy(
                requires=constants.AMPHORA,
                inject={constants.AMPHORA: failed_amp}))

        # Check that the VIP port exists and is ok
        failover_LB_flow.add(
            network_tasks.AllocateVIPforFailover(
                requires=constants.LOADBALANCER, provides=constants.VIP))

        # Update the database with the VIP information
        failover_LB_flow.add(database_tasks.UpdateVIPAfterAllocation(
            requires=(constants.LOADBALANCER_ID, constants.VIP),
            provides=constants.LOADBALANCER))

        # Make sure the SG has the correct rules and re-apply to the
        # VIP port. It is not used on the VIP port, but will help lock
        # the SG as in use.
        failover_LB_flow.add(network_tasks.UpdateVIPSecurityGroup(
            requires=constants.LOADBALANCER_ID, provides=constants.VIP_SG_ID))

        new_amp_role = constants.ROLE_STANDALONE
        if lb_topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            new_amp_role = constants.ROLE_BACKUP

        # Get a replacement amphora and plug all of the networking.
        #
        # Do this early as the compute services have been observed to be
        # unreliable. The community decided the chance that deleting first
        # would open resources for an instance is less likely than the compute
        # service failing to boot an instance for other reasons.
        if failed_amp:
            failed_vrrp_is_ipv6 = False
            if failed_amp.get(constants.VRRP_IP):
                failed_vrrp_is_ipv6 = utils.is_ipv6(
                    failed_amp[constants.VRRP_IP])
            failover_LB_flow.add(
                self.amp_flows.get_amphora_for_lb_failover_subflow(
                    prefix=constants.FAILOVER_LOADBALANCER_FLOW,
                    role=new_amp_role,
                    failed_amp_vrrp_port_id=failed_amp.get(
                        constants.VRRP_PORT_ID),
                    is_vrrp_ipv6=failed_vrrp_is_ipv6))
        else:
            failover_LB_flow.add(
                self.amp_flows.get_amphora_for_lb_failover_subflow(
                    prefix=constants.FAILOVER_LOADBALANCER_FLOW,
                    role=new_amp_role))

        if lb_topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            failover_LB_flow.add(database_tasks.MarkAmphoraBackupInDB(
                name=constants.MARK_AMP_BACKUP_INDB,
                requires=constants.AMPHORA))

        # Delete the failed amp
        if failed_amp:
            failover_LB_flow.add(
                self.amp_flows.get_delete_amphora_flow(failed_amp))

        # Update the data stored in the flow from the database
        failover_LB_flow.add(database_tasks.ReloadLoadBalancer(
            requires=constants.LOADBALANCER_ID,
            provides=constants.LOADBALANCER))

        # Configure the listener(s)
        # We will run update on this amphora again later if this is
        # an active/standby load balancer because we want this amp
        # functional as soon as possible. It must run again to update
        # the configurations for the new peers.
        failover_LB_flow.add(amphora_driver_tasks.AmpListenersUpdate(
            name=constants.AMP_LISTENER_UPDATE,
            requires=(constants.LOADBALANCER, constants.AMPHORA),
            inject={constants.TIMEOUT_DICT: timeout_dict}))

        # Bring up the new "backup" amphora VIP now to reduce the outage
        # on the final failover. This dropped the outage from 8-9 seconds
        # to less than one in my lab.
        # This does mean some steps have to be repeated later to reconfigure
        # for the second amphora as a peer.
        if lb_topology == constants.TOPOLOGY_ACTIVE_STANDBY:

            failover_LB_flow.add(database_tasks.CreateVRRPGroupForLB(
                name=new_amp_role + '-' + constants.CREATE_VRRP_GROUP_FOR_LB,
                requires=constants.LOADBALANCER_ID))

            failover_LB_flow.add(network_tasks.GetAmphoraNetworkConfigsByID(
                name=(new_amp_role + '-' +
                      constants.GET_AMPHORA_NETWORK_CONFIGS_BY_ID),
                requires=(constants.LOADBALANCER_ID, constants.AMPHORA_ID),
                provides=constants.FIRST_AMP_NETWORK_CONFIGS))

            failover_LB_flow.add(
                amphora_driver_tasks.AmphoraUpdateVRRPInterface(
                    name=new_amp_role + '-' + constants.AMP_UPDATE_VRRP_INTF,
                    requires=constants.AMPHORA,
                    inject={constants.TIMEOUT_DICT: timeout_dict},
                    provides=constants.FIRST_AMP_VRRP_INTERFACE))

            failover_LB_flow.add(amphora_driver_tasks.AmphoraVRRPUpdate(
                name=new_amp_role + '-' + constants.AMP_VRRP_UPDATE,
                requires=(constants.LOADBALANCER_ID, constants.AMPHORA),
                rebind={constants.AMPHORAE_NETWORK_CONFIG:
                        constants.FIRST_AMP_NETWORK_CONFIGS,
                        constants.AMP_VRRP_INT:
                        constants.FIRST_AMP_VRRP_INTERFACE},
                inject={constants.TIMEOUT_DICT: timeout_dict}))

            failover_LB_flow.add(amphora_driver_tasks.AmphoraVRRPStart(
                name=new_amp_role + '-' + constants.AMP_VRRP_START,
                requires=constants.AMPHORA,
                inject={constants.TIMEOUT_DICT: timeout_dict}))

            # Start the listener. This needs to be done here because
            # it will create the required haproxy check scripts for
            # the VRRP deployed above.
            # A "V" or newer amphora-agent will remove the need for this
            # task here.
            # TODO(johnsom) Remove this in the "X" cycle
            failover_LB_flow.add(amphora_driver_tasks.ListenersStart(
                name=new_amp_role + '-' + constants.AMP_LISTENER_START,
                requires=(constants.LOADBALANCER, constants.AMPHORA)))

            #  #### Work on standby amphora if needed #####

            new_amp_role = constants.ROLE_MASTER
            failed_amp = None
            if amps:
                failed_amp = amps.pop()

            if failed_amp:
                failed_amp_role = failed_amp.get(constants.ROLE)
                if failed_amp_role in (constants.ROLE_MASTER,
                                       constants.ROLE_BACKUP):
                    amp_role = 'master_or_backup'
                elif failed_amp_role == constants.ROLE_STANDALONE:
                    amp_role = 'standalone'
                elif failed_amp_role is None:
                    amp_role = 'spare'
                else:
                    amp_role = 'undefined'
                LOG.info("Performing failover for amphora: %s",
                         {"id": failed_amp.get(constants.ID),
                          "load_balancer_id": lb.get(constants.ID),
                          "lb_network_ip": failed_amp.get(
                              constants.LB_NETWORK_IP),
                          "compute_id": failed_amp.get(constants.COMPUTE_ID),
                          "role": amp_role})

                failover_LB_flow.add(
                    database_tasks.MarkAmphoraPendingDeleteInDB(
                        name=(new_amp_role + '-' +
                              constants.MARK_AMPHORA_PENDING_DELETE),
                        requires=constants.AMPHORA,
                        inject={constants.AMPHORA: failed_amp}))

                failover_LB_flow.add(database_tasks.MarkAmphoraHealthBusy(
                    name=(new_amp_role + '-' +
                          constants.MARK_AMPHORA_HEALTH_BUSY),
                    requires=constants.AMPHORA,
                    inject={constants.AMPHORA: failed_amp}))

            # Get a replacement amphora and plug all of the networking.
            #
            # Do this early as the compute services have been observed to be
            # unreliable. The community decided the chance that deleting first
            # would open resources for an instance is less likely than the
            # compute service failing to boot an instance for other reasons.
            failover_LB_flow.add(
                self.amp_flows.get_amphora_for_lb_failover_subflow(
                    prefix=(new_amp_role + '-' +
                            constants.FAILOVER_LOADBALANCER_FLOW),
                    role=new_amp_role))

            failover_LB_flow.add(database_tasks.MarkAmphoraMasterInDB(
                name=constants.MARK_AMP_MASTER_INDB,
                requires=constants.AMPHORA))

            # Delete the failed amp
            if failed_amp:
                failover_LB_flow.add(
                    self.amp_flows.get_delete_amphora_flow(
                        failed_amp))
                failover_LB_flow.add(
                    database_tasks.DisableAmphoraHealthMonitoring(
                        name=(new_amp_role + '-' +
                              constants.DISABLE_AMP_HEALTH_MONITORING),
                        requires=constants.AMPHORA,
                        inject={constants.AMPHORA: failed_amp}))

        # Remove any extraneous amphora
        # Note: This runs in all topology situations.
        #       It should run before the act/stdby final listener update so
        #       that we don't bother attempting to update dead amphorae.
        delete_extra_amps_flow = unordered_flow.Flow(
            constants.DELETE_EXTRA_AMPHORAE_FLOW)
        for amp in amps:
            LOG.debug('Found extraneous amphora %s on load balancer %s. '
                      'Deleting.', amp.get(constants.ID), lb.get(id))
            delete_extra_amps_flow.add(
                self.amp_flows.get_delete_amphora_flow(amp))

        failover_LB_flow.add(delete_extra_amps_flow)

        if lb_topology == constants.TOPOLOGY_ACTIVE_STANDBY:
            # Update the data stored in the flow from the database
            failover_LB_flow.add(database_tasks.ReloadLoadBalancer(
                name=new_amp_role + '-' + constants.RELOAD_LB_AFTER_AMP_ASSOC,
                requires=constants.LOADBALANCER_ID,
                provides=constants.LOADBALANCER))

            failover_LB_flow.add(database_tasks.GetAmphoraeFromLoadbalancer(
                name=new_amp_role + '-' + constants.GET_AMPHORAE_FROM_LB,
                requires=constants.LOADBALANCER_ID,
                provides=constants.AMPHORAE))

            # Listeners update needs to be run on all amphora to update
            # their peer configurations. So parallelize this with an
            # unordered subflow.
            update_amps_subflow = unordered_flow.Flow(
                constants.UPDATE_AMPS_SUBFLOW)

            # Setup parallel flows for each amp. We don't know the new amp
            # details at flow creation time, so setup a subflow for each
            # amp on the LB, they let the task index into a list of amps
            # to find the amphora it should work on.
            update_amps_subflow.add(
                amphora_driver_tasks.AmphoraIndexListenerUpdate(
                    name=(constants.AMPHORA + '-0-' +
                          constants.AMP_LISTENER_UPDATE),
                    requires=(constants.LOADBALANCER, constants.AMPHORAE),
                    inject={constants.AMPHORA_INDEX: 0,
                            constants.TIMEOUT_DICT: timeout_dict}))
            update_amps_subflow.add(
                amphora_driver_tasks.AmphoraIndexListenerUpdate(
                    name=(constants.AMPHORA + '-1-' +
                          constants.AMP_LISTENER_UPDATE),
                    requires=(constants.LOADBALANCER, constants.AMPHORAE),
                    inject={constants.AMPHORA_INDEX: 1,
                            constants.TIMEOUT_DICT: timeout_dict}))

            failover_LB_flow.add(update_amps_subflow)

            # Configure and enable keepalived in the amphora
            failover_LB_flow.add(self.amp_flows.get_vrrp_subflow(
                new_amp_role + '-' + constants.GET_VRRP_SUBFLOW,
                timeout_dict, create_vrrp_group=False))

            # #### End of standby ####

            # Reload the listener. This needs to be done here because
            # it will create the required haproxy check scripts for
            # the VRRP deployed above.
            # A "V" or newer amphora-agent will remove the need for this
            # task here.
            # TODO(johnsom) Remove this in the "X" cycle
            failover_LB_flow.add(
                amphora_driver_tasks.AmphoraIndexListenersReload(
                    name=(new_amp_role + '-' +
                          constants.AMPHORA_RELOAD_LISTENER),
                    requires=(constants.LOADBALANCER, constants.AMPHORAE),
                    inject={constants.AMPHORA_INDEX: 1,
                            constants.TIMEOUT_DICT: timeout_dict}))

        # Remove any extraneous ports
        # Note: Nova sometimes fails to delete ports attached to an instance.
        #       For example, if you create an LB with a listener, then
        #       'openstack server delete' the amphora, you will see the vrrp
        #       port attached to that instance will remain after the instance
        #       is deleted.
        # TODO(johnsom) Fix this as part of
        #               https://storyboard.openstack.org/#!/story/2007077

        # Mark LB ACTIVE
        failover_LB_flow.add(
            database_tasks.MarkLBActiveInDB(mark_subobjects=True,
                                            requires=constants.LOADBALANCER))

        return failover_LB_flow



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils

from octavia.volume import volume_base as driver_base

LOG = logging.getLogger(__name__)


class NoopManager(object):
    def __init__(self):
        super(NoopManager, self).__init__()
        self.volumeconfig = {}

    def create_volume_from_image(self, image_id):
        LOG.debug("Volume %s no-op, image id %s",
                  self.__class__.__name__, image_id)
        self.volumeconfig[image_id] = (image_id, 'create_volume_from_image')
        volume_id = uuidutils.generate_uuid()
        return volume_id

    def delete_volume(self, volume_id):
        LOG.debug("Volume %s no-op, volume id %s",
                  self.__class__.__name__, volume_id)
        self.volumeconfig[volume_id] = (volume_id, 'delete')

    def get_image_from_volume(self, volume_id):
        LOG.debug("Volume %s no-op, volume id %s",
                  self.__class__.__name__, volume_id)
        self.volumeconfig[volume_id] = (volume_id, 'get_image_from_volume')
        image_id = uuidutils.generate_uuid()
        return image_id


class NoopVolumeDriver(driver_base.VolumeBase):
    def __init__(self):
        super(NoopVolumeDriver, self).__init__()
        self.driver = NoopManager()

    def create_volume_from_image(self, image_id):
        volume_id = self.driver.create_volume_from_image(image_id)
        return volume_id

    def delete_volume(self, volume_id):
        self.driver.delete_volume(volume_id)

    def get_image_from_volume(self, volume_id):
        image_id = self.driver.get_image_from_volume(volume_id)
        return image_id



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import contextlib

from oslo_config import cfg
from oslo_log import log
from taskflow.jobs import backends as job_backends
from taskflow.persistence import backends as persistence_backends

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class JobboardTaskFlowDriver(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def job_board(self, persistence):
        """Setting up jobboard backend based on configuration setting.

        :param persistence: taskflow persistence backend instance
        :return: taskflow jobboard backend instance
        """


class MysqlPersistenceDriver(object):

    def __init__(self):
        self.persistence_conf = {
            'connection': CONF.task_flow.persistence_connection,
            'max_pool_size': CONF.database.max_pool_size,
            'max_overflow': CONF.database.max_overflow,
            'pool_timeout': CONF.database.pool_timeout,
        }

    def initialize(self):
        # Run migrations once on service start.
        backend = persistence_backends.fetch(self.persistence_conf)
        with contextlib.closing(backend):
            with contextlib.closing(backend.get_connection()) as connection:
                connection.upgrade()

    @contextlib.contextmanager
    def get_persistence(self):
        # Rewrite taskflow get backend, so it won't run migrations on each call
        backend = persistence_backends.fetch(self.persistence_conf)
        with contextlib.closing(backend):
            with contextlib.closing(backend.get_connection()) as conn:
                conn.validate()
            yield backend


class ZookeeperTaskFlowDriver(JobboardTaskFlowDriver):

    def __init__(self, persistence_driver):
        self.persistence_driver = persistence_driver

    def job_board(self, persistence):
        job_backends_hosts = ','.join(
            ['%s:%s' % (host, CONF.task_flow.jobboard_backend_port)
             for host in CONF.task_flow.jobboard_backend_hosts])
        jobboard_backend_conf = {
            'board': 'zookeeper',
            'hosts': job_backends_hosts,
            'path': '/' + CONF.task_flow.jobboard_backend_namespace,
        }
        jobboard_backend_conf.update(
            CONF.task_flow.jobboard_zookeeper_ssl_options)
        return job_backends.backend(CONF.task_flow.jobboard_name,
                                    jobboard_backend_conf,
                                    persistence=persistence)


class RedisTaskFlowDriver(JobboardTaskFlowDriver):

    def __init__(self, persistence_driver):
        self.persistence_driver = persistence_driver

    def job_board(self, persistence):
        jobboard_backend_conf = {
            'board': 'redis',
            'host': CONF.task_flow.jobboard_backend_hosts[0],
            'port': CONF.task_flow.jobboard_backend_port,
            'password': CONF.task_flow.jobboard_backend_password,
            'namespace': CONF.task_flow.jobboard_backend_namespace,
        }
        jobboard_backend_conf.update(
            CONF.task_flow.jobboard_redis_backend_ssl_options)
        return job_backends.backend(
            CONF.task_flow.jobboard_backend_namespace,
            jobboard_backend_conf,
            persistence=persistence)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from cryptography import fernet
from oslo_config import cfg
from stevedore import driver as stevedore_driver
from taskflow import task

from octavia.common import utils

CONF = cfg.CONF


class BaseCertTask(task.Task):
    """Base task to load drivers common to the tasks."""

    def __init__(self, **kwargs):
        super(BaseCertTask, self).__init__(**kwargs)
        self.cert_generator = stevedore_driver.DriverManager(
            namespace='octavia.cert_generator',
            name=CONF.certificates.cert_generator,
            invoke_on_load=True,
        ).driver


class GenerateServerPEMTask(BaseCertTask):
    """Create the server certs for the agent comm

    Use the amphora_id for the CN
    """

    def execute(self, amphora_id):
        cert = self.cert_generator.generate_cert_key_pair(
            cn=amphora_id,
            validity=CONF.certificates.cert_validity_time)
        key = utils.get_compatible_server_certs_key_passphrase()
        fer = fernet.Fernet(key)

        return fer.encrypt(cert.certificate + cert.private_key)



# Copyright 2019 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import multiprocessing

from octavia_lib.api.drivers import driver_lib as octavia_driver_lib
from octavia_lib.common import constants as lib_consts
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils
from stevedore import driver as stevedore_driver

from octavia.api.drivers.driver_agent import driver_listener
from octavia.common import config
from octavia.common import constants
from octavia.db import repositories
from octavia.tests.common import sample_certs
from octavia.tests.common import sample_data_models
from octavia.tests.functional.db import base

CONF = cfg.CONF


class DriverAgentTest(base.OctaviaDBTestBase):

    def _process_cleanup(self):
        self.exit_event.set()
        self.status_listener_proc.join(5)
        self.stats_listener_proc.join(5)
        self.get_listener_proc.join(5)

    def setUp(self):
        status_socket_file = '/tmp/octavia-{}.status.sock'.format(
            uuidutils.generate_uuid())
        stats_socket_file = '/tmp/octavia-{}.stats.sock'.format(
            uuidutils.generate_uuid())
        get_socket_file = '/tmp/octavia-{}.get.sock'.format(
            uuidutils.generate_uuid())
        sqlite_db_file = '/tmp/octavia-{}.sqlite.db'.format(
            uuidutils.generate_uuid())
        sqlite_db_connection = 'sqlite:///{}'.format(sqlite_db_file)

        # Note that because the driver agent is a multi-process
        # agent we must use a sqlite file rather than an
        # in-memory instance.
        super(DriverAgentTest, self).setUp(
            connection_string=sqlite_db_connection)

        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))
        conf.config(group="driver_agent",
                    status_socket_path=status_socket_file)
        conf.config(group="driver_agent",
                    stats_socket_path=stats_socket_file)
        conf.config(group="driver_agent", status_request_timeout=1)
        conf.config(group="driver_agent", get_socket_path=get_socket_file)
        conf.config(group="certificates", cert_manager='local_cert_manager')
        conf.config(group="certificates", storage_path='/tmp')

        # Set up the certificate
        cert_manager = stevedore_driver.DriverManager(
            namespace='octavia.cert_manager',
            name=CONF.certificates.cert_manager,
            invoke_on_load=True,
        ).driver
        self.cert_ref = cert_manager.store_cert(
            None,
            sample_certs.X509_CERT,
            sample_certs.X509_CERT_KEY_ENCRYPTED,
            sample_certs.X509_IMDS,
            private_key_passphrase=sample_certs.X509_CERT_KEY_PASSPHRASE)
        self.addCleanup(cert_manager.delete_cert, None, self.cert_ref)

        self.exit_event = multiprocessing.Event()

        self.status_listener_proc = multiprocessing.Process(
            name='status_listener', target=driver_listener.status_listener,
            args=(self.exit_event,))
        # TODO(johnsom) Remove once https://bugs.python.org/issue6721
        #               is resolved.
        self.status_listener_proc.daemon = True

        self.status_listener_proc.start()

        self.stats_listener_proc = multiprocessing.Process(
            name='stats_listener', target=driver_listener.stats_listener,
            args=(self.exit_event,))
        # TODO(johnsom) Remove once https://bugs.python.org/issue6721
        #               is resolved.
        self.stats_listener_proc.daemon = True

        self.stats_listener_proc.start()

        self.get_listener_proc = multiprocessing.Process(
            name='get_listener', target=driver_listener.get_listener,
            args=(self.exit_event,))
        # TODO(johnsom) Remove once https://bugs.python.org/issue6721
        #               is resolved.
        self.get_listener_proc.daemon = True

        self.get_listener_proc.start()

        self.addCleanup(self._process_cleanup)

        self.driver_lib = octavia_driver_lib.DriverLibrary(
            status_socket=status_socket_file,
            stats_socket=stats_socket_file,
            get_socket=get_socket_file)

        self.sample_data = sample_data_models.SampleDriverDataModels()
        self.repos = repositories.Repositories()

        # Create the full load balancer in the database
        self.tls_container_dict = {
            lib_consts.CERTIFICATE: sample_certs.X509_CERT.decode('utf-8'),
            lib_consts.ID: sample_certs.X509_CERT_SHA1,
            lib_consts.INTERMEDIATES: [
                i.decode('utf-8') for i in sample_certs.X509_IMDS_LIST],
            lib_consts.PASSPHRASE: None,
            lib_consts.PRIMARY_CN: sample_certs.X509_CERT_CN,
            lib_consts.PRIVATE_KEY: sample_certs.X509_CERT_KEY.decode('utf-8')}

        # ### Create load balancer
        self.repos.flavor_profile.create(
            self.session, id=self.sample_data.flavor_profile_id,
            provider_name=constants.AMPHORA,
            flavor_data='{"something": "else"}')
        self.repos.flavor.create(
            self.session, id=self.sample_data.flavor_id,
            enabled=True, flavor_profile_id=self.sample_data.flavor_profile_id)
        self.repos.create_load_balancer_and_vip(
            self.session, self.sample_data.test_loadbalancer1_dict,
            self.sample_data.test_vip_dict)

        # ### Create Pool
        pool_dict = copy.deepcopy(self.sample_data.test_pool1_dict)

        pool_dict[constants.LOAD_BALANCER_ID] = self.sample_data.lb_id

        # Use a live certificate
        pool_dict[constants.TLS_CERTIFICATE_ID] = self.cert_ref
        pool_dict[constants.CA_TLS_CERTIFICATE_ID] = self.cert_ref
        pool_dict[constants.CRL_CONTAINER_ID] = self.cert_ref

        # Remove items that are linked in the DB
        del pool_dict[lib_consts.MEMBERS]
        del pool_dict[constants.HEALTH_MONITOR]
        del pool_dict[lib_consts.SESSION_PERSISTENCE]
        del pool_dict[lib_consts.LISTENERS]
        del pool_dict[lib_consts.L7POLICIES]

        self.repos.pool.create(self.session, **pool_dict)

        self.repos.session_persistence.create(
            self.session, pool_id=self.sample_data.pool1_id,
            type=lib_consts.SESSION_PERSISTENCE_SOURCE_IP)

        self.provider_pool_dict = copy.deepcopy(
            self.sample_data.provider_pool1_dict)
        self.provider_pool_dict[
            constants.LISTENER_ID] = self.sample_data.listener1_id

        # Fix for render_unsets = True
        self.provider_pool_dict[
            lib_consts.SESSION_PERSISTENCE][lib_consts.COOKIE_NAME] = None
        self.provider_pool_dict[lib_consts.SESSION_PERSISTENCE][
            lib_consts.PERSISTENCE_GRANULARITY] = None
        self.provider_pool_dict[lib_consts.SESSION_PERSISTENCE][
            lib_consts.PERSISTENCE_TIMEOUT] = None

        # Use a live certificate
        self.provider_pool_dict[
            lib_consts.TLS_CONTAINER_DATA] = self.tls_container_dict
        self.provider_pool_dict[lib_consts.TLS_CONTAINER_REF] = self.cert_ref
        self.provider_pool_dict[
            lib_consts.CA_TLS_CONTAINER_DATA] = (
                sample_certs.X509_CERT.decode('utf-8'))
        self.provider_pool_dict[
            lib_consts.CA_TLS_CONTAINER_REF] = self.cert_ref
        self.provider_pool_dict[
            lib_consts.CRL_CONTAINER_DATA] = (
                sample_certs.X509_CERT.decode('utf-8'))
        self.provider_pool_dict[lib_consts.CRL_CONTAINER_REF] = self.cert_ref

        # ### Create Member
        member_dict = copy.deepcopy(self.sample_data.test_member1_dict)
        self.repos.member.create(self.session, **member_dict)
        self.provider_pool_dict[lib_consts.MEMBERS] = [
            self.sample_data.provider_member1_dict]

        # ### Create Health Monitor
        hm_dict = copy.deepcopy(self.sample_data.test_hm1_dict)
        self.repos.health_monitor.create(self.session, **hm_dict)
        self.provider_pool_dict[
            lib_consts.HEALTHMONITOR] = self.sample_data.provider_hm1_dict

        # ### Create Listener
        listener_dict = copy.deepcopy(self.sample_data.test_listener1_dict)
        listener_dict[lib_consts.DEFAULT_POOL_ID] = self.sample_data.pool1_id

        # Remove items that are linked in the DB
        del listener_dict[lib_consts.L7POLICIES]
        del listener_dict[lib_consts.DEFAULT_POOL]
        del listener_dict[constants.SNI_CONTAINERS]

        # Use a live certificate
        listener_dict[constants.TLS_CERTIFICATE_ID] = self.cert_ref
        listener_dict[constants.CLIENT_CA_TLS_CERTIFICATE_ID] = self.cert_ref
        listener_dict[constants.CLIENT_CRL_CONTAINER_ID] = self.cert_ref

        self.repos.listener.create(self.session,
                                   **listener_dict)
        self.repos.sni.create(self.session,
                              listener_id=self.sample_data.listener1_id,
                              tls_container_id=self.cert_ref, position=1)

        # Add our live certs in that differ from the fake certs in sample_data
        self.provider_listener_dict = copy.deepcopy(
            self.sample_data.provider_listener1_dict)
        self.provider_listener_dict[
            lib_consts.DEFAULT_TLS_CONTAINER_REF] = self.cert_ref
        self.provider_listener_dict[
            lib_consts.DEFAULT_TLS_CONTAINER_DATA] = self.tls_container_dict
        self.provider_listener_dict[
            lib_consts.CLIENT_CA_TLS_CONTAINER_REF] = self.cert_ref
        self.provider_listener_dict[
            lib_consts.CLIENT_CA_TLS_CONTAINER_DATA] = (
                sample_certs.X509_CERT.decode('utf-8'))
        self.provider_listener_dict[
            lib_consts.CLIENT_CRL_CONTAINER_REF] = self.cert_ref
        self.provider_listener_dict[
            lib_consts.CLIENT_CRL_CONTAINER_DATA] = (
                sample_certs.X509_CERT.decode('utf-8'))
        self.provider_listener_dict[
            lib_consts.SNI_CONTAINER_DATA] = [self.tls_container_dict]
        self.provider_listener_dict[
            lib_consts.SNI_CONTAINER_REFS] = [self.cert_ref]

        self.provider_listener_dict[
            lib_consts.DEFAULT_POOL] = self.provider_pool_dict
        self.provider_listener_dict[
            lib_consts.DEFAULT_POOL_ID] = self.sample_data.pool1_id

        self.provider_listener_dict[lib_consts.L7POLICIES] = [
            self.sample_data.provider_l7policy1_dict]

        # ### Create L7 Policy
        l7policy_dict = copy.deepcopy(self.sample_data.test_l7policy1_dict)
        del l7policy_dict[lib_consts.L7RULES]
        self.repos.l7policy.create(self.session, **l7policy_dict)

        # ### Create L7 Rules
        l7rule_dict = copy.deepcopy(self.sample_data.test_l7rule1_dict)
        self.repos.l7rule.create(self.session, **l7rule_dict)
        l7rule2_dict = copy.deepcopy(self.sample_data.test_l7rule2_dict)
        self.repos.l7rule.create(self.session, **l7rule2_dict)

        self.provider_lb_dict = copy.deepcopy(
            self.sample_data.provider_loadbalancer_tree_dict)
        self.provider_lb_dict[lib_consts.POOLS] = [self.provider_pool_dict]
        self.provider_lb_dict[
            lib_consts.LISTENERS] = [self.provider_listener_dict]

    def test_get_loadbalancer(self):
        result = self.driver_lib.get_loadbalancer(self.sample_data.lb_id)

        self.assertEqual(self.provider_lb_dict,
                         result.to_dict(render_unsets=True, recurse=True))

        # Test non-existent load balancer
        result = self.driver_lib.get_loadbalancer('bogus')
        self.assertIsNone(result)

    def test_get_listener(self):
        result = self.driver_lib.get_listener(self.sample_data.listener1_id)

        # We need to recurse here to pick up the SNI data
        self.assertEqual(self.provider_listener_dict,
                         result.to_dict(render_unsets=True, recurse=True))

        # Test non-existent listener
        result = self.driver_lib.get_listener('bogus')
        self.assertIsNone(result)

    def test_get_pool(self):
        result = self.driver_lib.get_pool(self.sample_data.pool1_id)

        self.assertEqual(self.provider_pool_dict,
                         result.to_dict(render_unsets=True, recurse=True))

        # Test non-existent pool
        result = self.driver_lib.get_pool('bogus')
        self.assertIsNone(result)

    def test_get_member(self):
        result = self.driver_lib.get_member(self.sample_data.member1_id)

        self.assertEqual(self.sample_data.provider_member1_dict,
                         result.to_dict(render_unsets=True))

        # Test non-existent member
        result = self.driver_lib.get_member('bogus')
        self.assertIsNone(result)

    def test_get_healthmonitor(self):
        result = self.driver_lib.get_healthmonitor(self.sample_data.hm1_id)

        self.assertEqual(self.sample_data.provider_hm1_dict,
                         result.to_dict(render_unsets=True))

        # Test non-existent health monitor
        result = self.driver_lib.get_healthmonitor('bogus')
        self.assertIsNone(result)

    def test_get_l7policy(self):
        result = self.driver_lib.get_l7policy(self.sample_data.l7policy1_id)

        self.assertEqual(self.sample_data.provider_l7policy1_dict,
                         result.to_dict(render_unsets=True, recurse=True))

        # Test non-existent L7 policy
        result = self.driver_lib.get_l7policy('bogus')
        self.assertIsNone(result)

    def test_get_l7rule(self):
        result = self.driver_lib.get_l7rule(self.sample_data.l7rule1_id)

        self.assertEqual(self.sample_data.provider_l7rule1_dict,
                         result.to_dict(render_unsets=True))

        # Test non-existent L7 rule
        result = self.driver_lib.get_l7rule('bogus')
        self.assertIsNone(result)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from octavia.amphorae.backends.utils import network_utils
from octavia.common import exceptions
from octavia.tests.common import sample_network_data
import octavia.tests.unit.base as base


class TestNetworkUtils(base.TestCase):

    def setUp(self):
        super(TestNetworkUtils, self).setUp()

    def test_find_interface(self):
        FAKE_INTERFACE = 'fake0'
        IPV4_ADDRESS = '203.0.113.55'
        BROADCAST_ADDRESS = '203.0.113.55'
        IPV6_ADDRESS = '2001:db8::55'
        SAMPLE_IPV4_ADDR = sample_network_data.create_iproute_ipv4_address(
            IPV4_ADDRESS, BROADCAST_ADDRESS, FAKE_INTERFACE)
        SAMPLE_IPV6_ADDR = sample_network_data.create_iproute_ipv6_address(
            IPV6_ADDRESS, FAKE_INTERFACE)
        SAMPLE_INTERFACE = sample_network_data.create_iproute_interface(
            FAKE_INTERFACE)
        BROKEN_INTERFACE = [{'attrs': []}]

        mock_ip_addr = mock.MagicMock()
        mock_rtnl_api = mock.MagicMock()
        mock_rtnl_api.get_addr.side_effect = [[], SAMPLE_IPV4_ADDR,
                                              SAMPLE_IPV6_ADDR,
                                              SAMPLE_IPV6_ADDR]
        mock_rtnl_api.get_links.side_effect = [SAMPLE_INTERFACE,
                                               SAMPLE_INTERFACE,
                                               BROKEN_INTERFACE]

        # Test no match
        IPV4_ADDRESS = '203.0.113.55'
        mock_ip_addr.version = 4

        self.assertIsNone(network_utils._find_interface(IPV4_ADDRESS,
                                                        mock_rtnl_api,
                                                        IPV4_ADDRESS))

        # Test with IPv4 address
        mock_rtnl_api.reset_mock()
        mock_ip_addr.version = 4

        result = network_utils._find_interface(IPV4_ADDRESS, mock_rtnl_api,
                                               IPV4_ADDRESS)

        self.assertEqual(FAKE_INTERFACE, result)
        mock_rtnl_api.get_addr.assert_called_once_with(address=IPV4_ADDRESS)
        mock_rtnl_api.get_links.assert_called_once_with(2)

        # Test with IPv6 address
        mock_rtnl_api.reset_mock()
        mock_ip_addr.version = 6

        result = network_utils._find_interface(IPV6_ADDRESS, mock_rtnl_api,
                                               IPV6_ADDRESS)

        self.assertEqual(FAKE_INTERFACE, result)
        mock_rtnl_api.get_addr.assert_called_once_with(address=IPV6_ADDRESS)
        mock_rtnl_api.get_links.assert_called_once_with(2)

        # Test with a broken interface
        mock_rtnl_api.reset_mock()
        mock_ip_addr.version = 6

        self.assertIsNone(network_utils._find_interface(IPV6_ADDRESS,
                                                        mock_rtnl_api,
                                                        IPV6_ADDRESS))
        mock_rtnl_api.get_addr.assert_called_once_with(address=IPV6_ADDRESS)
        mock_rtnl_api.get_links.assert_called_once_with(2)

    @mock.patch('octavia.amphorae.backends.utils.network_utils.'
                '_find_interface')
    @mock.patch('pyroute2.IPRoute', create=True)
    @mock.patch('pyroute2.NetNS', create=True)
    def test_get_interface_name(self, mock_netns, mock_ipr, mock_find_int):
        FAKE_INTERFACE = 'fake0'
        FAKE_NETNS = 'fake-ns'
        IPV4_ADDRESS = '203.0.113.64'

        mock_ipr_enter_obj = mock.MagicMock()
        mock_ipr_obj = mock.MagicMock()
        mock_ipr_obj.__enter__.return_value = mock_ipr_enter_obj
        mock_ipr.return_value = mock_ipr_obj

        mock_netns_enter_obj = mock.MagicMock()
        mock_netns_obj = mock.MagicMock()
        mock_netns_obj.__enter__.return_value = mock_netns_enter_obj
        mock_netns.return_value = mock_netns_obj

        mock_find_int.side_effect = [FAKE_INTERFACE, FAKE_INTERFACE, None]

        # Test a bogus IP address
        self.assertRaises(exceptions.InvalidIPAddress,
                          network_utils.get_interface_name, 'not an IP', None)

        # Test with no network namespace
        result = network_utils.get_interface_name(IPV4_ADDRESS)

        self.assertEqual(FAKE_INTERFACE, result)
        mock_ipr.assert_called_once_with()
        mock_find_int.assert_called_once_with(IPV4_ADDRESS, mock_ipr_enter_obj,
                                              IPV4_ADDRESS)

        # Test with network namespace
        mock_ipr.reset_mock()
        mock_find_int.reset_mock()

        result = network_utils.get_interface_name(IPV4_ADDRESS,
                                                  net_ns=FAKE_NETNS)
        self.assertEqual(FAKE_INTERFACE, result)
        mock_netns.assert_called_once_with(FAKE_NETNS)
        mock_find_int.assert_called_once_with(IPV4_ADDRESS,
                                              mock_netns_enter_obj,
                                              IPV4_ADDRESS)

        # Test no interface found
        mock_ipr.reset_mock()
        mock_find_int.reset_mock()

        self.assertRaises(
            exceptions.NotFound, network_utils.get_interface_name,
            IPV4_ADDRESS, net_ns=FAKE_NETNS)
